var version = "1.0.10";
//1.0.08 - added SCC export
//1.0.07 - added CEA608Decoder and SCC import
//1.0.06 - heartbeat loss from client warns the user
//1.0.05 - pushing cadet.config to configdir area (Win==%appdata/cadet/cadet.config%, MacOS==~/"Library/Application Support/cadet/cadet.config"


//Timecode types
var ZEROS = "00:00:00.00";
var INFINITY = "24:00:00.00";
var HIDDEN_TIME = "99:99:99.99";
var MIN_BLINK_TIME = 0.01; //default interval between outtime and next intime, in seconds

//Alignments
var ALIGN_LEFT = 1;
var ALIGN_CENTER = 2;
var ALIGN_RIGHT = 3;
var ALIGN_TOP = 1;
var ALIGN_BOTTOM = 2;
var CEA_STYLE_POPON = 1;
var CEA_STYLE_ROLLUP = 2;
var CEA_STYLE_PAINTON = 3;
var CEA_DEFAULT_STYLE = CEA_STYLE_POPON;
var CEA_DEFAULT_DEPTH = 3;

var CONFIG_DIR = __dirname;
var CONFIG_FILE = "cadet.config";

/// global configuration
var fs = require('fs');
var url = require('url');
var mime = require('mime');
var himalaya = require('himalaya');//for reading ttml as html
var extend = require('extend');//for cloning an object (in JS, direct assignment is by reference)
var open = require('open');



///////  START FILEBROWSER DECLARES //////
var http = require('http');
var _ = require('lodash');
var express = require('express');
var fs = require('fs');
var path = require('path');
var util = require('util');
var os = require('os');

var program = require('commander');

var log = initializeLogger("info");

log.info("SERVER SESSION DATE = ["+new Date().toISOString().replace(/T/, ' ').replace(/\..+/, '')+"]");
log.info(versionInfo());
log.info("NodeJS process PID=["+process.pid+"]");


function collect(val, memo) {
  if(val && val.indexOf('.') != 0) val = "." + val;
  memo.push(val);
  return memo;
}


program
  .option('-p, --port <port>', 'Port to run the file-browser. Default value is 8000')
  .option('-e, --exclude <exclude>', 'File extensions to exclude. To exclude multiple extension pass -e multiple times. e.g. ( -e .js -e .cs -e .swp) ', collect, [])
  .option('-d, --dir <dir>', 'Home directory')
  .option('-c, --configdir <dir>', 'Config file directory')
  .parse(process.argv);


///////  END FILEBROWSER DECLARES //////


/// command line arguments
var optimist = require("optimist");
var argv = optimist
    .options({
        port: {
            description: "TCP Port",
            alias: "p",
            default: 8000,
        },
        configdir: {
            description: "Config dir",
            alias: "c",
            default: "",
        },
        reserve: {
            describe: "reserve a channel- double quotes surround channel and commas separate components",
            alias: "r"
        },
        logLevel: {
            describe: "loggin level - error, info, connection, debug",
            alias: ["l", "log"],
            default: "info"
        },
        sslPort: {
            default: 0
        },
        help: {
            describe: "show this help message",
            alias: "h"
        }
    }).argv;

if (argv["help"]) {
    optimist.showHelp();
    process.exit(1);
} // if


var port = argv["port"];
//console.log("using port=["+port+"]");
//var sslPort = argv["sslPort"];
//var reservedChannels = argv["reserve"] || "";
//console.log("PORT from args=["+port+"]");

//set up the config file dir/path
var CONFIG_DIR = "";//argv["configdir"];
if( program.configdir )
	CONFIG_DIR = program.configdir
else
	CONFIG_DIR = __dirname;
log.info("CONFIG_DIR=["+CONFIG_DIR+"]");
var CONFIG_PATH = CONFIG_DIR + "/" + CONFIG_FILE
log.info("CONFIG_PATH=["+CONFIG_PATH+"]");



var webroot=""; //= program.dir.toString().substr(1);
//console.log("SERVER: parsing commandline: webroot=["+webroot+"]");

if( !loadConfig() ){
	//if( program.dir.toString().trim().length > 0 ){
	if( program.dir ){	//give priority to command-line-supplied --dir
		//console.log("program.dir = ["+program.dir+"]");
		//webroot = program.dir.toString().substr(1);
		webroot = program.dir;
		//console.log("!loadConfig, program.dir exists: webroot=["+webroot+"]");
	} else {	//otherwise just use root, last chance
		webroot = "/";
		program.dir = webroot;
		//console.log("!loadConfig, NO program.dir exists: webroot=["+webroot+"]");
	}
}


dir = webroot;
program.dir = webroot;


var serverStatisticsUpdate = null; // used to cancel setTimeout

var _ = require("underscore");
var express = require("express");
var app = express();

/*
var bodyParser = require('body-parser');
app.use(bodyParser.json()); // support json encoded bodies
app.use(bodyParser.urlencoded({ extended: true })); // support encoded bodies
*/
var bodyParser = require('body-parser');
app.use(bodyParser.json({limit: '50mb'}));
app.use(bodyParser.urlencoded({limit: '50mb', extended: true}));


var http = require("http");
var server = http.createServer(app);

Number.prototype.toTimecode = function (n, comma) {
	if( this == null || this == -1 ) return " ";
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	var sec_num = this;
	var hours   = Math.floor(sec_num / 3600);
	var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	var seconds = (sec_num - (hours * 3600) - (minutes * 60)).toFixed(n);

	if (hours   < 10) {hours   = "0"+hours;}
	if (minutes < 10) {minutes = "0"+minutes;}
	if (seconds < 10) {seconds = "0"+seconds;}
	var time    = hours+':'+minutes+':'+seconds;
	if( comma )
		time = time.replace(".",",");
	return time;
}
Number.prototype.toTimecodeSMPTE = function () {
	if( this == -1 ) return " ";
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	var sec_num = this;
	var hours   = Math.floor(sec_num / 3600);
	var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
	var seconds = Math.floor(sec_num - (hours * 3600) - (minutes * 60));
	//var frames = sec_num - (hours * 3600) - (minutes * 60) - (seconds * 60);
	var frames = Math.floor((sec_num % 1) * 30);

	if (hours   < 10) {hours   = "0"+hours;}
	if (minutes < 10) {minutes = "0"+minutes;}
	if (seconds < 10) {seconds = "0"+seconds;}
	if (frames < 10) {frames = "0"+frames;}
	var time    = hours+':'+minutes+':'+seconds+';'+frames;
	return time;
}
Number.prototype.toHexString = function () {
	if( this < 0 || this > 255 ) return "??";
	var temp = this.toString(16);
	//console.log("toHexString: temp=["+temp+"]");
	//var sfoo = ("00" + temp).substring(2 - temp.length, 2);
	var sfoo = ("00" + temp).substr(temp.length, 2);
	return sfoo;
}
//allow for malformed timecodes, like h:mm:ss.sss  (h/m/s should have leading zeros
String.prototype.timecodeToNumber = function () {
				if( this.trim().length==0 )	//blank timecode represented as -1 numerically
					return -1;
			
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	//01234567890
	//hh:mm:ss.ss
	var parts = this.split(":");
	var sec_num = parseFloat(parts[2]);
	var hours   = parseFloat(parts[0])*3600;
	var minutes = parseFloat(parts[1])*60;
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	return hours+minutes+sec_num;
}
String.prototype.timecodeToNumberOld = function () {
				if( this.trim().length==0 )	//blank timecode represented as -1 numerically
					return -1;
			
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	//01234567890
	//hh:mm:ss.ss
	var sec_num = parseFloat(this.substr(6,this.length));
	var hours   = parseFloat(this.substr(0,2))*3600;
	var minutes = parseFloat(this.substr(3,2))*60;
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	return hours+minutes+sec_num;
}
String.prototype.timecodeSmpteToNumber = function () {
	if( this.trim().length==0 )	//blank timecode represented as -1 numerically
		return -1;
			
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	//01234567890
	//hh:mm:ss;ff
	var hours   = parseFloat(this.substr(0,2))*3600;
	var minutes = parseFloat(this.substr(3,2))*60;
	var secs = parseFloat(this.substr(6,2));
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	var frames = parseFloat(this.substr(9,2));
	var secs_total = secs+(frames/30.0);
	//console.log("timecodeSmpteToNumber: in=["+this+"] hours=["+hours+"] minutes=["+minutes+"] secs=["+secs+"] frames=["+frames+"] secs_total=["+secs_total+"]");
	return hours+minutes+secs_total;
}



//log.info("01:02:3.5=["+"01:02:3.5".timecodeToNumber()+"] 1:02:3.5=["+"1:02:3.5".timecodeToNumber()+"] ");


//delete all .temp.* files from /public/clips directory

server.listen(port, function() {
    log.info("Server listening on port " + server.address().port);
    log.info("File browsering root: " + webroot);
}).on('error', function(){	//assume CADET is already running on the port
	log.info("Error listening on port:"+port);
	return;
});
open("http://localhost:"+port+"/");
//FOR LATER, LAUNCH SPECIFIC BROWSER [need to check OS, since Chrome = "Google Chrome" (MAC) "chrome"(WIN) ""google-chrome"(Linux)]
//var cp = require("child_process");
//cp.exec('open -a "Google Chrome" http://www.robbot.com', null); // notice this without a callback..

//////////////////////////////////////
// start HEARTBEAT code: checks for the last client-side heartbeat (coming from interval in client that POSTs to this server
var heartbeatTimer;
var HEARTBEAT_CHECK_DELAY = 70; //10 second pause in heartbeats suggest client is gone
heartbeatTimer = setInterval(heartbeatCheck,1000);
var lastHeartbeatTime;
var warnedNoHeartbeat = false;
function heartbeatCheck(){

	//initialize  lastHeartbeatTime
	if( !lastHeartbeatTime ){
		lastHeartbeatTime = new Date();
	}
		
	var currHeartbeatTime = new Date();
	
	var timeDiff = currHeartbeatTime-lastHeartbeatTime;
	// strip the ms
	timeDiff /= 1000;
	// get seconds (Original had 'round' which incorrectly counts 0:28, 0:29, 1:30 ... 1:59, 1:0)
	var seconds;
	seconds = timeDiff; //Math.round(timeDiff % 60);

	if( seconds > HEARTBEAT_CHECK_DELAY ){
		if( !warnedNoHeartbeat ){
    		log.info("!!!!!! NO HEARTBEAT FROM BROWSER, YOU SHOULD PROBABLY QUIT APPLICATION !!!!!!"); 
    		warnedNoHeartbeat = true;
    	}
    	//quitServer();
	} else {
    	//log.debug("heart beating");
    }
}
// end HEARBEAT code
//////////////////////////////////////



/// routes

//////////////////////////
// ADD HEADERS
//////////////////////////
app.use(function (req, res, next) {

    // Website you wish to allow to connect
    res.setHeader('Access-Control-Allow-Origin', 'http://localhost:8000');
    res.setHeader('Access-Control-Allow-Origin', 'http://127.0.0.1:8000');
    res.setHeader('Access-Control-Allow-Origin', 'http://localhost');
    res.setHeader('Access-Control-Allow-Origin', 'http://127.0.0.1');

    // Request methods you wish to allow
    res.setHeader('Access-Control-Allow-Methods', 'GET, POST, OPTIONS, PUT, PATCH, DELETE');

    // Request headers you wish to allow
    res.setHeader('Access-Control-Allow-Headers', 'X-Requested-With,content-type');

    // Set to true if you need the website to include cookies in the requests sent
    // to the API (e.g. in case you use sessions)
    res.setHeader('Access-Control-Allow-Credentials', true);

    // Pass to next layer of middleware
    next();
});


//////////////////////////
// MAIN ENTRY FILE ROUTES
//////////////////////////
app.get('/', function(req, res){
    //log.debug("app.get /:  sending cadet.html ");
    res.sendFile('cadet.html', {"root": __dirname} );
});
app.get("/cadet", function(req, res, next) {
    //log.debug("routes:all: /cadet");
    res.sendFile("cadet.html", {"root": __dirname});
}); // app.get
app.get("/cadet.html", function(req, res, next) {
    //log.debug("routes:all: /cadet.html");
    res.sendFile("cadet.html", {"root": __dirname});
}); // app.get
app.post("/quit/", function(req, res, next) {
    //log.debug("/quit/ called");
    log.info("**CADET QUIT BY USER**");
    //server.close();
    //process.exit(0)
    quitServer();
}); // app.post /quit/
app.post("/clientVersion/", function(req, res, next) {
    //log.debug("/clientVersion/ called");
    //log.debug("req.body="+JSON.stringify(req.body));
    var myObject = JSON.parse(req.body.thedata);
    log.info("Browser reports client version: v"+myObject.clientVersion);
    res.status(200).send("Browser reports client version: v"+myObject.clientVersion);
}); // app.post /quit/
/*************
****************/

function quitServer(){
    server.close();
    process.exit(0)
}//quitServer

////////////////////////////////
// ROUTE CONFIGEDIT 
////////////////////////////////
app.get("/configedit/:rand", function(req, res) {
	//log.debug("app.get(//configedit//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in

	var thedata = sendConfigHtml(res);	
	//log.debug("app.post(//configedit//): thedata=["+thedata+"]");
    
	//res.send("SERVER:configedit cadet.config=["+thedata+"]");
	return;
});//app.post("/configedit/"

//config.html is set up to allow user to edit cadet configuration
function sendConfigHtml(res){
	//log.debug("sendConfigHtml: enter");
	
	//log.debug("sendConfigHtml: __dirname=["+__dirname+"]");
	var thefile = __dirname+"/config.html";
	//log.debug("sendConfigHtml: thefile=["+thefile+"]");
		
	try{
		fs.readFile(thefile, 'utf8', function (err,data) {
		//log.debug("sendConfigHtml: data=["+data+"]");
		if( data ){
			data = data.replace("%%webroot%%",webroot);
			res.send( data );
		} else {
			//log.debug("sendConfigHtml: Error reading file=["+thefile+"]  data=undefined");
			res.status(200).send({});
			return log.debug("sendConfigHtml: Error reading file=["+thefile+"]  data=undefined");
		}	
		});
	} catch(err){
		log.debug("sendConfigHtml: Error reading file=["+thefile+"]");
		res.status(200).send("config doesn't exist");
		return log.debug(err);
	}
}//sendConfigHtml`

// CONFIG-SAVE
app.post("/configsave/", function(req, res) {
	//log.debug("app.post(//configsave//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in

	var myObject = JSON.parse(req.body.configdefaults);
	//log.debug("app.post(//configsave//): configdefaults.defaults.webroot=["+JSON.stringify(myObject.defaults.webroot)+"]");
	if( !dirExists(myObject.defaults.webroot) ){
		log.debug("app.post(//configsave//): webroot=["+myObject.defaults.webroot+"] doesn't exist, returning");
		res.status(400).send("Server says webroot dir=["+myObject.defaults.webroot+"] doesn't exist");
		return;
	}
    
	//var wstream = fs.createWriteStream(__dirname + "/" + CONFIG_FILE);
	//var wstream = fs.createWriteStream(CONFIG_DIR + "/" + CONFIG_FILE);
	var wstream;

	//log.debug("first try to create directory=["+CONFIG_DIR+"]");
	var thefile = CONFIG_DIR + "/" + CONFIG_FILE
	ensureDirectoryExistence(thefile)
	//log.debug("creating writeStream for ["+thefile+"]");
	wstream = fs.createWriteStream(thefile);
		
	
	wstream.write(JSON.stringify(myObject,null,2));
	wstream.end();
	
	//save it locally
	webroot = myObject.defaults.webroot;
	dir = webroot;//??
	res.status(200).send("OK");
	return;
});//app.post("/configsave/"
//config.html is set up to allow user to edit cadet configuration

function loadConfig(){
	//try to load webroot from config file, if possible; return true if successful
	//log.debug("loadConfig: enter");

	//var thefile = __dirname + "/" + CONFIG_FILE;
	var thefile = CONFIG_DIR + "/" + CONFIG_FILE;
	//log.debug("loadConfig: thefile=["+thefile+"]");

	try{
		data = fs.readFileSync(thefile, "utf8");
		var obj = JSON.parse(data);
		//log.debug("loadConfig: from file json webroot=["+obj.defaults.webroot+"]");
		if( obj.defaults && obj.defaults.webroot ){
			//console.log("loadConfig: setting webroot=["+obj.defaults.webroot+"]");
			webroot = obj.defaults.webroot.toString();
		} else {
			//console.log("loadConfig: obj.defaults doesn't exist or obj.defaults.webroot doesn't exist");
		}

		//console.log("loadConfig: finally webroot=["+webroot+"]");
		return true;
	} catch (err){
		log.info("NO CONFIG FILE FOUND, USING DEFAULTS");
		return false;
	}
}//loadConfig


// CONFIG-READ
app.post("/configread/", function(req, res) {
	//log.debug("app.post(//configread//): ENTER");
	res.send( readConfig() );
	return;
});//app.post("/configread/"
function readConfig(){
	//log.debug("readConfig: enter");

	//var thefile = __dirname + "/" + CONFIG_FILE;
	var thefile = CONFIG_DIR + "/" + CONFIG_FILE;
	//log.debug("readConfig: thefile=["+thefile+"]");

	try{
		
		data = fs.readFileSync(thefile, "utf8");
		var obj = JSON.parse(data);
		//console.log("readConfig: from file json webroot=["+obj.defaults.webroot+"]");
		if( obj.defaults ){
			//console.log("readConfig: returning raw data");
			return data;
		} else {
			//console.log("readConfig: obj.defaults doesn't exist");
			return null;
		}

	} catch (err){
		log.info("NO CONFIG FILE FOUND, USING DEFAULTS");
		return null;
	}
}//readConfig

////////////////////////////////
// ROUTE GETWEBROOT
////////////////////////////////
app.get("/getwebroot/", function(req, res) {
	//log.debug("app.post(//configsave//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
	//log.debug("app.get(/getwebroot/: webroot=["+webroot+"]");
	var obj = {};
	obj.webroot = webroot;
	res.send(JSON.stringify(obj));
});//app.post("/getwebroot/"


////////////////////////////////
// ROUTE CADET SUPPORT FILES 
////////////////////////////////
app.get("/-cad3t-/*", function(req, res, next) {
    //log.debug("app.get /-cad3t-/*: req=[" + req.params[0] +"]");
    
    var thefile = req.params[0];//.replace("-cad3t-/","");

    try{ 
    	//log.debug("app.get /-cad3t-/*: trying root=["+__dirname+"] thefile=["+thefile+"]");
    	fs.statSync(__dirname + "/" + thefile);
    	//log.debug("app.get /-cad3t-/*: IS A FILE IN ./");
    	if( thefile == "config.html" ){
    	} else {
    		res.sendFile(thefile, {"root": __dirname});
    	}
    }
    catch(ex){
				res.send("Error: " + req.params[0]+ " not found");
    }// catch ./
    //next();
}); // app.get


//////////////////////////
// ROUTE FILEBROWSER
//////////////////////////
var includeMap=null;
app.get('/files', function(req, res) {
 //log.debug("APP.GET /files: dir=["+dir+"]");
 if( req.query.includeMap ){
	 includeMap = JSON.parse(req.query.includeMap);
	 //log.debug("APP.GET /files: includeMap=["+JSON.stringify(includeMap)+"]");
	 //for(var i=0; i<includeMap.length; i++)
	 	//log.debug("includeMap["+i+"]=["+includeMap[i]+"]");
 }
 var oldCurrentDir =  dir;
 var currentDir =  dir;
 currentDir =  dir;
 //log.debug("app.get /files: oldCurrentDir=["+oldCurrentDir+"]");
 var query = req.query.path || '';
 //log.debug("app.get /files: desiredDir=["+query+"]");
 if (query) currentDir = path.join(dir, query);
 //log.debug("app.get /files: currentDir=["+currentDir+"]");
 //log.debug("app.get /files: browsing currentDir=["+currentDir+"]");
 try{
 	currentDir = currentDir;
	 //fixup symbolic links like "My Documents" in Windows
	 currentDir = fs.realpathSync(currentDir)
 
	 fs.readdir(currentDir, function (err, files) {
		 if (err) {
			console.log("Error trying to scan directory = ["+currentDir+"]");
			//throw err;
			currentDir = oldCurrentDir;
			return;
		  }
		  //console.log("files:  fs.readdir returned files.length=["+files.length+"]");
		  var data = [];
		  files
		  .filter(function (file) {
			  return true;
		  }).forEach(function (file) {
			try {
					//console.log("/files/: processing file=["+file+"]");
				
					//ignore hidden/special files on mac
					if( file.startsWith(".") || file.startsWith("$") ){
						return;
					}
				  
					var isDirectory = fs.statSync(path.join(currentDir,file)).isDirectory();
					if (isDirectory) {
					  //console.log("/files: isDirectory=true file=["+file+"]");
					  data.push({ Name : file, IsDirectory: true, Path : path.join(query, file)  });
					} else {
					  //console.log("/files: isDirectory=false file=["+file+"]");
					  //check for excludes
					  var ext = path.extname(file);
					  //console.log("/files: isDirectory=false ext=["+ext+"]");
					  if(program.exclude && _.contains(program.exclude, ext)) {
						//console.log("excluding file ", file);
						return;
					  }
					  //check for includes 

					  if( includeMap[0]=="*" ){
						//console.log("/files: include due to *");
						data.push({ Name : file, Ext : ext, IsDirectory: false, Path : path.join(query, file) });
					  }else if( ext && _.contains(includeMap, ext) ){
						//console.log("/files: contains includeMap");
						data.push({ Name : file, Ext : ext, IsDirectory: false, Path : path.join(query, file) });
					  }
					}

			} catch(e) {
			  console.log(e); 
			}        
		
		  });
		  data = _.sortBy(data, function(f) { return f.Name });
		  res.json(data);
		  //console.log("APP.GET /files: jsondata=["+JSON.stringify(data)+"]");
	  });
  }//try watching for errors like inaccessible directories like "My Documents" in Windows
  catch(e) {
  	console.log("app.get /files: Error on readdir for dir=["+currentDir+"]");
  }
});


//////////////////////////
// ROUTE checkExists
//////////////////////////
app.post('/checkExists/', function(req, res) {
    //log.debug(".post(/checkExists/): ENTER");
    //log.debug(".post(/checkExists/): ENTER data passed=["+JSON.stringify(req.body,null,2)+"]");
	var myObj = JSON.parse(req.body.thedata);
    //log.debug(".post(/checkExists/): ENTER to check on ["+myObj.src+"]");
    
	try {
    	//log.debug(".post(/checkExists/): trying ["+webroot+myObj.src+"]");
		var stats = fs.statSync(webroot+myObj.src);
    	//log.debug(".post(/checkExists/): statStync says exists ["+myObj.src+"]");
		res.send("Server says ["+myObj.src+"] exists");
	}
	catch(err) {
    	//log.debug(".post(/checkExists/): statStync says doesn't exist ["+myObj.src+"]");
		res.status(400).send("Server says ["+myObj.src+"] doesn't exist");
	}
    
});//app.post('/checkExists/:src'
function dirExists(path){
	return(fs.existsSync(path));
}//dirExists
function ensureDirectoryExistence(thepath){
	//log.debug("ensureDirectoryExistence: ENTER with path=["+thepath+"]");
	var thedirname = path.dirname(thepath);
	if (dirExists(thedirname)){
		//log.debug("ensureDirectoryExistence: EXISTS already path=["+thedirname+"]");
		return true;
	} else {
		//log.debug("ensureDirectoryExistence: doesn't exist yet, mkdirSync path=["+thedirname+"]");
		fs.mkdirSync(thedirname);
		return true;
	}
	return ensureDirectoryExistence(dirname);
}//ensureDirectoryExistence
function ensureDirectoryExistenceOrig(path){
	var dirname = path.dirname(filePath);
	if (fs.existsSync(dirname)) {
		return true;
	}
	ensureDirectoryExistence(dirname);
	fs.mkdirSync(dirname);
}//ensureDirectoryExistenceOrig

///////////////////////////
// ALL OTHER FILE ROUTES
///////////////////////////
app.get("/*", function(req, res, next) {
    //log.debug("app.get /*: req.params[0]=[" + req.params[0] +"]");
    
    try{ 
    	//log.debug("app.get /*: trying webroot/req.params[0]=[" + webroot + "/" + req.params[0]+"]");
    	fs.statSync(webroot + "/" + req.params[0]);
    	//log.debug("app.get /*: IS A FILE DIRECTLY");
    	res.status(200).sendFile(req.params[0], {"root": webroot + "/"});
    }
    catch(ex){
				//log.debug("app.get /*: ["+webroot + "/" + req.params[0]+"] IS NOT A FILE");
				res.send("Error: file=[" + webroot + "/" +req.params[0]+ "] not found");
    }// catch 
    //next();
}); // app.get

///////////////////////////
// ROUTE HEARTBEAT
///////////////////////////
app.post("/heartbeat/", function(req, res) {
	heartbeat();
	res.send("200");
});//app.post("/heartbeat/"
var lastHeartbeatTime = null;
function heartbeat(){
    	//log.debug("heartbeat, updating the lastHeartbeatTime");
    	//update the lastHeartbeatTime with current time. There's a timer running that's checking the time since last updated... QUITs if too long
    	lastHeartbeatTime = new Date();
    	if( warnedNoHeartbeat ){
    		log.info("Heart beating, browser connection is alive!");
    	}
    	warnedNoHeartbeat = false;
};//heartbeat


///////////////////////////
// ROUTE SAVE
///////////////////////////
app.post("/saveproject/", function(req, res) {
	//log.debug(".post(/saveproject/): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);
    //log.debug(".post(/saveproject/*): stringify myObject=["+JSON.stringify(myObject)+"]");
    
    //log.debug(".post(/saveproject/*): myObject.mediaFilename=["+myObject.mediaFilename+"]");
    //log.debug(".post(/saveproject/*): myObject.trackFilename=["+myObject.trackFilename+"]");
    //log.debug(".post(/saveproject/*): events.length=["+myObject.events.length+"]");
    //log.debug(".post(/saveproject/*): stringify element 0._text=["+JSON.stringify(myObject.events[0], null, 2)+"]");
    //log.debug(".post(/saveproject/*): element 0.text=["+myObject.events[0].text+"]");
    //log.debug(".post(/saveproject/*): element 0.htmlText=["+myObject.events[0].htmlText+"]");
    
    
    //save the project file
    //saveProject("myProject.cadet", myObject);
    saveProject(myObject, res);
    
    //res.send("Saved project");
    return;
    
});//app.post("/saveproject/"
function saveProject(oProject, res){
	//log.debug("saveProject: enter");

    var projName, projDir, projPath;
    
    if( oProject.config.webProjectFile.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectFILE setting to Untitled.cadet");
    	projName = "Untitled.cadet";
    } else {
    	//log.debug("saveProject: yes json webprojectFILE=["+oProject.config.webProjectFile+"]");
    	projName = oProject.config.webProjectFile;    		
    }
    //fix it up to end with .cadet
	if( !projName.endsWith(".cadet") )
		projName += ".cadet";    		

    	
    //if( oProject.config.webProjectDir.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectDIR setting to webroot");
    	//projDir = webroot;
    //} else {
    	//log.debug("saveProject: yes json webprojectDIR=["+oProject.config.webProjectDir+"]");
    	projDir = oProject.config.webProjectDir;
    //}
    	
    //if( oProject.config.webProjectPath.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectPATH setting to webroot+'/'+projName");
    	//projPath = webroot+"/"+ projDir + "/"+projName;
    	//projPath = webroot  + projDir + "/"+ projName; //webroot always ends with slash??
    	if( !webroot.endsWith("/") && !projDir.startsWith("/") && !projDir.startsWith("\\") )
    		projPath = webroot + "/" + projDir + "/"+ projName; //webroot always ends with slash??
    	else
    		projPath = webroot + projDir + "/"+ projName; //webroot always ends with slash??
    	log.info("saveProject: PATH=["+projPath+"]");
    	
    //} else {
    	//log.debug("saveProject: yes json webprojectPATH=["+oProject.config.webProjectPath+"]");
    	//projPath = oProject.config.webProjectPath;
    //}
    	
	//projFullPath = oProject.config.mediaDir+"/"+sname
    
    //log.debug("saveProject: oProject=["+JSON.stringify(oProject, null, 2)+"]");
    //log.debug("saveProject: projPath=["+projPath+"]");
	var strm = fs.createWriteStream(projPath);
	strm.write(JSON.stringify(oProject, null, 2));
	strm.end
	
	res.send("Saved project file to ["+projPath+"]");
}//saveProject

///////////////////////////
// ROUTE REFRESHCAPTIONS
///////////////////////////
app.post("/refreshCaptions*", function(req, res) {
	//log.debug("app.post(//refreshCaptions//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);  
    
    if(myObject.config.mediaFilename.trim().length==0){
    	//log.debug("post(/refreshCaptions*: NO BASEMEDIA=["+myObject.config.mediaFilename+"]");
    	res.send("No base media file at["+myObject.config.mediaFilename+"]");
    	return;
    }
    
 	//log.debug("root __dirname=["+__dirname+"]");
	//log.debug("app.post(//refreshCaptions//): clip path=["+myObject.config.mediaDir+"/"+myObject.config.mediaFilename+"]");
    //log.debug("app.post(//refreshCaptions//): trak path=["+myObject.config.mediaDir+"/"+myObject.config.trackFilename+"]");
	//return;//!!!!!!!!!!!!!!!
	
	
    //write a VTT file using media path as the base path for track; it will always live next to the media
	
    var trackFullPath;
    
    //get base path (all except .extention) but start at homeDir/public
    var basemediapath, basemediadir;
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    basemediadir = webroot+myObject.config.mediaDir + "/";
    basemediapath = webroot+myObject.config.mediaDir + "/" + myObject.config.mediaFilename;
    basemediapath = basemediapath.substring(0, basemediapath.lastIndexOf(".")); 
    
    //write the temp file
    trackFullPath = basemediapath+".temp.vtt";
    
    //make sure basemediapath exists
    if( !fs.existsSync(basemediadir) ){
    	log.debug("post(/refreshCaptions*: NOFIND=["+basemediadir+"]");
    	res.send("No media file at["+basemediadir+"]");
    	return;
    }
    
    
    //log.debug(".post(//refreshCaptions//): writing temp VTT file=["+trackFullPath+"]");
	writeVTT(trackFullPath,myObject);
    
	res.send("Caption file refreshed");
	return;
});//app.post("/refreshCaptions/"

//////////////////////////
// ROUTE EXPORTTYPE
//////////////////////////
app.post("/exporttype/:etype", function(req, res) {
	//log.debug("app.post(//exporttype//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);    
    
 	//log.debug("root __dirname=["+__dirname+"]");
	//log.debug("app.post(//exporttype//): clip path=["+myObject.config.mediaDir+"/"+myObject.config.mediaFilename+"]");
    //log.debug("app.post(//exporttype//): trak path=["+myObject.config.mediaDir+"/"+myObject.config.trackFilename+"]");
    

    //get base path (all except .extention)
    var exportFilename, exportDir, exportPath;
    var basemediapath;
    
    if( myObject.config.exportFilename && myObject.config.exportDir ){
    	exportFilename = myObject.config.exportFilename;
    	exportDir = myObject.config.exportDir;
    } else if( myObject.config.mediaFilename && myObject.config.mediaDir ){
    	exportFilename = myObject.config.mediaFilename;
    	exportDir = myObject.config.mediaDir;
    } else {
    	exportFilename = "Untitled."
    	exportDir = "";
    }
    //convert all backslash to forwardslash
    exportDir = exportDir.replace(/\\/g, '/');
    //log.debug("//exporttype//: webroot=["+webroot+"] exportDir=["+exportDir+"] exportFilename=["+exportFilename+"]");
    
    if( exportDir == "/" ) exportDir = "";
    //exportPath = webroot + "/" + exportDir + "/" + exportFilename;
    //strip leading/trailing slashes and insert them as necessary
    //exportPath = "/" + webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;
	//WINDOWS TEMPLATE
	if( webroot.substr(1,1)==":" ) //don't prepend root slash
    	exportPath = webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;
	else
    	exportPath = "/" + webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;


    exportPath = exportPath.substring(0, exportPath.lastIndexOf(".")); 
    //log.debug("//export//: exportPath=["+exportPath+"]");
    
    /*
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    if( !myObject.config.mediaFilename )
    	myObject.config.mediaFilename = "Untitled.";
    basemediapath = webroot+myObject.config.mediaDir + "/" + myObject.config.mediaFilename;
    basemediapath = basemediapath.substring(0, basemediapath.lastIndexOf(".")); 
    //log.debug("app.post(//exporttype:etype//): basemediapath=["+basemediapath+"]");
    
    
    exportPath = basemediapath;
    */
    
    //log.debug("app.post(//exporttype//): export type=["+req.params.etype+"]");
    var etype = req.params.etype.toString().toLowerCase();
    var trackFullPath;
	switch (etype.toLowerCase()) {
		case "vtt":
			trackFullPath = exportPath+".vtt"; //basemediapath+".vtt";
			//log.debug("app.post(//exporttype:etype//): writing VTT file=["+trackFullPath+"]");
			writeVTT(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "srt":
			trackFullPath = exportPath+".srt";
			//log.debug("app.post(//exporttype:etype//): writing SRT file=["+trackFullPath+"]");
			writeSRT(trackFullPath,myObject);
			var thefile = exportFilename;
			thefile = exportFilename.substring(0, exportFilename.lastIndexOf("."));
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>******************<br/>Note: if you intend to upload to Facebook, rename the file to '"+thefile+".en_US.srt'<br/>******************<br/>");
			return;
			break;
		case "ttml":
			trackFullPath = exportPath+".ttml";
			//log.debug("app.post(//exporttype:etype//): writing TTML file=["+trackFullPath+"]");
			writeTTML(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>******************<br/>Note: if you intend to view in IE or Edge, make sure to set the fileType on the server to 'application/ttml+xml'<br/>******************<br/>");
			return;
			break;
		case "qttext":
			trackFullPath = exportPath+".qttext.txt";
			//log.debug("app.post(//exporttype:etype//): writing VTT file=["+trackFullPath+"]");
			writeQTtext(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "sami":
			//console.log("export SAMI not implemented");
			trackFullPath = exportPath+".smi";
			//log.debug("app.post(//exporttype:etype//): writing SAMI file=["+trackFullPath+"]");
			writeSAMI(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "scc":
			//console.log("export SCC not implemented");
			trackFullPath = exportPath+".scc";
			//log.debug("app.post(//exporttype:etype//): writing SAMI file=["+trackFullPath+"]");
			writeSCC(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "plain":
			trackFullPath = exportPath+".txt";
			//log.debug("app.post(//exporttype:etype//): writing PLAIN TEXT file=["+trackFullPath+"]");
			writePlainText(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "ad":
			trackFullPath = exportPath+".html";
			//log.debug("app.post(//exporttype:etype//): writing VTT file=["+trackFullPath+"]");
			writeADText(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		default:
			console.log("No valid export type specified.");
			res.send("Invalid export type specified ["+etype+"]");
			return
	}
    
    //should never get here
	res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>");
	return;
});//app.post("/exporttype/"


////////////////
//// END REST
////////////////



/// utility functions



/// versioning

function versionInfo() {
    var path = require("path");
    return path.basename(__filename) + " v" + (arguments[0] || version);
} // versionInfo

/// Logging

function initializeLogger(logLevel) {
    var winston = require("winston");
    var logger = new(winston.Logger)({
        transports: [
                //new (winston.transports.File)({ filename: 'somefile.log' }),
                new(winston.transports.Console)({
                    level: logLevel
                })
            ] // transports
    }); // logger

    logger.setLevels({
        debug: 0, // noisy
        connection: 1, // connect/disconnect messages
        info: 2,
        error: 3 // only errors
    });
    return logger;
} // initializeLogger

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
// FILE EXPORT functions
///////////////////////
// EXPORT QTTEXT
///////////////////////
function writeQTtext(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	writeFileQTtext(wstream,obj)
	wstream.end();
}//writeQTtext
function writeFileQTtext(strm, o){
/*
https://developer.apple.com/library/mac/documentation/QuickTime/QT6_3/Chap1/QT6WhatsNew.html
{QTtext}{font:Geneva}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{timeStamps:absolute}{language:0}{textEncoding:0}
[00:00:00.000]
Trip to Drumlin Farm
[00:00:02.000]
{child noise}
[00:00:04.000]
Hi kids!
[00:00:06.000]
*/

	//write header section
	strm.write('{QTtext}\n');
	strm.write('{font:Arial}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{width:160}{height:48}{timeStamps:absolute}{language:0}{textEncoding:0}\n');
	strm.write('\n');
	
	var intime, outtime;
	for( var i=0; i<o.events.length; i++ ){
		intime = o.events[i].intime;
		outtime = o.events[i].outtime
		strm.write('\n');
		strm.write("["+o.events[i].intime.toTimecode(3) + ']\n');
		strm.write(eventToQTtext(o.events[i]) +'\n');
	}
	//write last outtime
	strm.write('\n');
	strm.write("["+outtime.toTimecode(3) + ']\n');
	
}//writeFileQTtext
function eventToQTtext(e){
	var sfoo="", ret="", tr={};
	for( var i=0; i<e.textRuns.length; i++ ){
		tr=e.textRuns[i];
		sfoo = tr.text;
		//log.debug("eventToQTtext: sfoo=["+sfoo+"]");
		if( sfoo.length==0 ) sfoo = " ";
		
		//prepend newline for this run
		if( tr.linebreak ) sfoo = "\n"; //sfoo = "\n"+sfoo;

		ret += sfoo;
	}//textRuns
	
	return ret;
}//eventToQTtext




///////////////////////
// EXPORT VTT
///////////////////////
function writeVTT(fpath, obj){
	var wstream = fs.createWriteStream(fpath, {encoding: "utf8"});
	writeFileVTT(wstream,obj)
	wstream.end();
}//writeVTT
function writeFileVTT(strm, o){
	//log.debug("writeFileVTT: data=["+JSON.stringify(o, null, 2)+"]");
	//write header section
	strm.write('WEBVTT\n');
	strm.write('\n');
	
	//maybe don't do this for IE???
	/*
	strm.write('STYLE\n');
	strm.write('\n');
	strm.write('::cue(.red){ color: red; }\n');
	strm.write('::cue(.blue){ color: blue; }\n');
	strm.write('::cue(.green){ color: green; }\n');
	strm.write('::cue(.yellow){ color: yellow; }\n');
	strm.write('::cue(.cyan){ color: cyan; }\n');
	strm.write('::cue(.magenta){ color: magenta; }\n');
	strm.write('::cue(.bgblack){ background-color: black; }\n');
	strm.write('::cue(.bgwhite){ background-color: white; }\n');
	strm.write('::cue(.bgred){ background-color: red; }\n');
	strm.write('::cue(.bgblue){ background-color: blue; }\n');
	strm.write('::cue(.bggreen){ background-color: green; }\n');
	strm.write('::cue(.bgyellow){ background-color: yellow; }\n');
	strm.write('::cue(.bgcyan){ background-color: cyan; }\n');
	strm.write('::cue(.bgmagenta){ background-color: magenta; }\n');
	strm.write('\n');
	*/
	
	for( var i=0; i<o.events.length; i++ ){
		strm.write('\n');
		strm.write((i+1)+'\n');
		strm.write(o.events[i].intime.toTimecode(3) + ' --> ' + o.events[i].outtime.toTimecode(3) +'  align:'+itoeAlign(o.events[i].align)+'  line:'+itoeVAlign(o.events[i].valign)+'\n');
		strm.write(eventToVTT(o.events[i]) +'\n');
	}
	strm.write('\n');
}//writeFileVTT
function eventToVTT(e){
	var sfoo="", ret="", tr={}, bgottext = false;
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text;
		
		/*//done on client side, in cleanUpBlankLines()
		if( i==0 ){	//first run, left trim
			sfoo = sfoo.replace(/^\s+/,"");
		}
		*/
		
		//log.debug("eventToVTT:1 tr.text=["+sfoo+"]");
		
		if( sfoo.trim().length > 0 )
			bgottext = true;
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		
		if( tr.linebreak ){
			if( bgottext ){	//linebreak and already found text
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=true");
				ret += "\n";
				bgottext = false;
				continue;
			} else {		//linebreak and no text yet
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=false");
				continue;
			}
		} else {
		}
		
				
		//log.debug("eventToVTT: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo.trim(); //trim the last trailing spaces
		//log.debug("eventToVTT:2 sfoo=["+sfoo+"]");
		if( tr.bold ) sfoo = sfoo.engulfVTT("b");
		if( tr.italic ) sfoo = sfoo.engulfVTT("i");
		if( tr.underline ) sfoo = sfoo.engulfVTT("u");
		sfoo = sfoo.engulfColorVTT(tr.color?tr.color:"white",tr.bgcolor?tr.bgcolor:"black");
		//log.debug("eventToVTT:3 sfoo=["+sfoo+"]");
		
		ret += sfoo;
		//log.debug("eventToVTT: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToVTT: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToVTT
String.prototype.engulfVTT = function (tag) {
	//wrap the string in <tag></tag>
	var sfoo = this;
	return("<"+tag+">"+this+"</"+tag+">"); 
}
String.prototype.engulfColorVTT = function (fgc,bgc) {
	//wrap the string in <c.fgc.bgc></c>
	var sfoo = this;
	var fgname=rgbToName(fgc), bgname=rgbToName(bgc);
	return("<c."+rgbToName(fgc)+".bg"+rgbToName(bgc)+">"+this+"</c>"); 
}


///////////////////////
// EXPORT SRT
///////////////////////
function writeSRT(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	writeFileSRT(wstream,obj)
	wstream.end();
}//writeSRT
function writeFileSRT(strm, o){
	//no header
	
	for( var i=0; i<o.events.length; i++ ){
		strm.write((i+1)+'\n'); //don't put in subtitle numbers for SRT, youtube doesn't like them
		strm.write(o.events[i].intime.toTimecode(3,true) + ' --> ' + o.events[i].outtime.toTimecode(3,true)+'\n');
		strm.write(eventToSRT(o.events[i]) +'\n');
		strm.write('\n');
	}
	strm.write('\n');
}//writeFileSRT
function eventToSRT(e){
	var sfoo="", ret="", tr={}, bgottext = false;
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text;
		
		if( sfoo.trim().length > 0 )
			bgottext = true;
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			if( bgottext ){	//linebreak and already found text
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=true");
				ret += "\n";
				bgottext = false;
				continue;
			} else {		//linebreak and no text yet
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=false");
				continue;
			}
		} else {
		}
		
		//log.debug("eventToSRT: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		
		ret += sfoo;
		//log.debug("eventToSRT: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToSRT: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToSRT


///////////////////////
// EXPORT TTML
///////////////////////
/*
<?xml version="1.0" encoding="utf-8"?>
<tt xml:lang="en"
    xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling">
  <head>
    <styling>
      <style id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />
    </styling>
  </head>
  <body id="thebody" style="defaultCaption">
    <div xml:lang="en">
      <p begin="00:00:07.36" end="00:00:09.63">All:<br/>What, Arthur, what is it?</p>
      <p begin="00:00:17.35" end="00:00:18.30" tts:textAlign="left">Ahem...</p>
      <p begin="00:00:18.30" end="00:00:19.35" tts:textAlign="right">What's the matter?</p>
    </div>
  </body>
</tt>
*/
function writeTTML(fpath, obj){
	//log.debug("writeTTML: enter");
	var wstream = fs.createWriteStream(fpath, {encoding: "utf8"});
	writeFileTTML(wstream,obj)
	wstream.end();
}//writeTTML
function writeFileTTML(strm, o){
	//log.debug("writeFileTTML: enter");
	//begin header
	strm.write('<?xml version="1.0" encoding="utf-8"?>' + os.EOL);
	strm.write('<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">' + os.EOL);
	strm.write('\t<head>' + os.EOL);
	strm.write('\t\t<styling>' + os.EOL);
	strm.write('\t\t\t<style xml:id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />' + os.EOL);
	strm.write('\t\t</styling>' + os.EOL);
	strm.write('\t</head>' + os.EOL);
	strm.write('\t<body xml:id="thebody" style="defaultCaption">' + os.EOL);
	strm.write('\t\t<div xml:lang="en">' + os.EOL);

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//set alignment
		//log.debug("writeFileTTML: align=["+ev.align+"]");
		switch (ev.align) {
			case 1:
				thisalign = "left"
				break;
			case 2:
				thisalign = "center"
				break;
			case 3:
				thisalign = "right"
				break;
			default:
				thisalign = "center"
		}//switch align
		
		//write timestamps
		strm.write('\t\t\t<p xml:id="s_'+(i+1)+'" begin="'+o.events[i].intime.toTimecode(2)+'" end="'+o.events[i].outtime.toTimecode(2)+'" tts:textAlign="'+thisalign+'">');
		
		//write textruns
		strm.write(eventToTTML(ev) + '</p>' + os.EOL);
	}
	
	strm.write('\t\t</div>' + os.EOL);
	strm.write('\t</body>' + os.EOL);
	strm.write('</tt>' + os.EOL);
	strm.write(os.EOL);

}//writeFileTTML
function eventToTTML(e){
	//log.debug("eventToTTML: enter");
	var sfoo="", ret="", tr={}, thelinetext="";
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToTTML: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak && thelinetext.trim().length > 0){
			ret += ('<br/>');
			thelinetext = "";
			continue;
		} else {
			thelinetext += tr.text;
		}
		
		//log.debug("eventToTTML: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		if( thelinetext.trim().length > 0 ){
			sfoo = sfoo.engulfTTML(tr);
			ret += sfoo;
		}
		//log.debug("eventToTTML: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToTTML: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToTTML
String.prototype.engulfTTML = function (tr) {
	//wrap the string in <c.fgc.bgc></c>
	var ret = this;
	var stts="";
	
	if( tr.bold ){
		stts += ' tts:fontWeight="bold"';
	}
	if( tr.italic ){
		stts += ' tts:fontStyle="italic"';
	}
	if( tr.underline ){
		stts += ' tts:textDecoration="underline"';
	}
	if( tr.color ){
		stts += (' tts:color="'+rgbToName(tr.color)+'"');
	}
	if( tr.bgcolor ){
		stts += (' tts:backgroundColor="'+rgbToName(tr.bgcolor)+'"');
	}
	
	if( stts.length > 0 )
		ret = '<span'+stts+'>' + this + '</span>';
		
	return( ret ); 
}//String.prototype.engulfTTML


///////////////////////
// EXPORT SAMI
///////////////////////
/*
<SAMI>
<HEAD>
   <STYLE TYPE = "text/css">
   <!--
   // P defines the basic style selector for closed caption paragraph text //
   P {font-family:sans-serif; color:white;}
   // Source, Small, and Big define additional ID selectors for closed caption text //
   #Source {color: orange; font-family: arial; font-size: 12pt;}
   #Small {Name: SmallTxt; font-size: 8pt; color: yellow;}
   #Big {Name: BigTxt; font-size: 12pt; color: magenta;}
   // ENUSCC and FRFRCC define language class selectors for closed caption text //
   .ENUSCC {Name: 'English Captions'; lang: en-US; SAMIType: CC;}
   .FRFRCC {Name: 'French Captions'; lang: fr-FR; SAMIType: CC;}
   -->
   </STYLE>
</HEAD>
<BODY>
   <!<entity type="mdash"/>- The closed caption text displays at 1000 milliseconds. -->
   <SYNC Start = 1000>
      <!-- English closed captions -->
      <P Class = ENUSCC ID = Source>Narrator
      <P Class = ENUSCC>Great reason to visit Seattle, brought to you by two out-of-staters.
      <!-- French closed captions -->
      <P Class = FRFRCC ID = Source>Narrateur
      <P Class = FRFRCC>Deux personnes ne venant la r&eacute;gion vous donnent de bonnes raisons de visiter Seattle.
</BODY>
</SAMI>
*/
function writeSAMI(fpath, obj){
	//log.debug("writeSAMI: enter");
	var wstream = fs.createWriteStream(fpath, {encoding: "utf8"});
	writeFileSAMI(wstream,obj)
	wstream.end();
}//writeSAMI
function writeFileSAMI(strm, o){
	//log.debug("writeFileSAMI: enter");
	//begin header
	strm.write('<SAMI>' + os.EOL);
	strm.write('\t<HEAD>' + os.EOL);
	strm.write("\t\t<STYLE TYPE='text/css'>" + os.EOL);
	strm.write("\t\t\t<!--" + os.EOL);
	strm.write("\t\t\tP {font-family:sans-serif; color:white; text-align:center;}" + os.EOL);
	strm.write("\t\t\t.ENUSCC {Name: 'English Captions'; lang: en-US; SAMIType: CC;}" + os.EOL);
	strm.write("\t\t\t-->" + os.EOL);
	strm.write('\t\t</StylE>' + os.EOL);
	strm.write('\t</HEAD>' + os.EOL);
	strm.write('\t<BODY>' + os.EOL);

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//set alignment
		//log.debug("writeFileSAMI: align=["+ev.align+"]");
		switch (ev.align) {
			case 1:
				thisalign = "left"
				break;
			case 2:
				thisalign = "center"
				break;
			case 3:
				thisalign = "right"
				break;
			default:
				thisalign = "center"
		}//switch align
		
		//write timestamps
		strm.write('\t\t<SYNC Start='+Math.floor(o.events[i].intime * 1000)+'>' + os.EOL);
		strm.write('\t\t\t<P Class=ENUSCC>');
		
		//write textruns
		strm.write(eventToSAMI(ev) + '</P>' + os.EOL);
		strm.write('\t\t</SYNC>' + os.EOL);
		
		//write the outtime as a separate caption
		strm.write('\t\t<SYNC Start='+Math.floor(o.events[i].outtime * 1000)+'>' + os.EOL);
		strm.write('\t\t\t<P Class=ENUSCC>&nbsp;<P>' + os.EOL);
		strm.write('\t\t</SYNC>' + os.EOL);

	}//each event
	
	strm.write('\t</BODY>' + os.EOL);
	strm.write('</SAMI>' + os.EOL);
	strm.write(os.EOL);

}//writeFileSAMI
function eventToSAMI(e){
	//log.debug("eventToTTML: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToTTML: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += ('<br/>');
			continue;
		} else {
		}
		
		//log.debug("eventToTTML: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		sfoo = sfoo.engulfSAMI(tr);
		
		ret += sfoo;
		//log.debug("eventToTTML: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToTTML: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToSAMI
String.prototype.engulfSAMI = function (tr) {
	//wrap the string in <c.fgc.bgc></c>
	var ret = this;
	
	return( ret );
	
	var stts="";
	
	if( tr.bold ){
		stts += ' tts:fontWeight="bold"';
	}
	if( tr.italic ){
		stts += ' tts:fontStyle="italic"';
	}
	if( tr.underline ){
		stts += ' tts:textDecoration="underline"';
	}
	if( tr.color ){
		stts += (' tts:color="'+rgbToName(tr.color)+'"');
	}
	if( tr.bgcolor ){
		stts += (' tts:backgroundColor="'+rgbToName(tr.bgcolor)+'"');
	}
	
	if( stts.length > 0 )
		ret = '<span'+stts+'>' + this + '</span>';
		
	return( ret ); 
}//String.prototype.engulfSAMI



///////////////////////
// EXPORT SCC
///////////////////////
function itoeCeaStyle(i){
	switch(i){
		case 1:
			return "POPON";
			break;
		case 2:
			return "ROLLUP";
			break;
		case 3:
			return "PAINTON";
			break;
		default:
			return "UNKNOWN CEA STYLE";
	}
}//itoeCeaStyle

function writeSCC(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	
	//console.log("writeSCC: obj=*******"+JSON.stringify(obj)+"*******");
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	//console.log("writeSCC:data=*******"+JSON.stringify(data)+"*******");
	
	/*
	console.log("");
	console.log("writeSCC: extended data=*******"+JSON.stringify(newobj)+"*******");
	var newobj = extend(true, data.events, obj.events);
	console.log("");
	console.log("writeSCC: extended data=*******"+JSON.stringify(newobj)+"*******");
	*/
	
	//create the proper CadetData object from the raw object data
	for(var e=0; e<obj.events.length; e++){
		var ev = obj.events[e];
		data.addEvent(ev.intime, ev.outtime, ev.align, ev.valign);
		for( var r=0; r<ev.textRuns.length; r++ ){
			run = ev.textRuns[r];
			var cadetStyle = new CadetStyle(run.bold, run.italic, run.underline, run.color, run.bgcolor, run.linebreak); 
			var textRun = new CadetTextRun( cadetStyle, run.text );
			data.addFullRun(textRun);
		}
	}
	//transfer style and depth info if it exists
	data.style = CEA_DEFAULT_STYLE;
	data.depth = CEA_DEFAULT_DEPTH;
	if( obj.style ) data.style = obj.style;
	if( obj.depth ) data.depth = obj.depth;
	//console.log("writeSCC: obj.style=["+itoeCeaStyle(obj.style)+"] obj.depth=["+obj.depth+"]");
	
	/*
	console.log("");
	console.log("writeSCC: data=*******"+JSON.stringify(data)+"*******");
	console.log("");
	console.log("writeSCC: align=["+data.getEvents()[0].align+"] valign=["+data.getEvents()[0].valign+"] linecount=["+data.getLines(0).length+"] *******"+JSON.stringify(data.getLines(0))+"*******");
	*/
	
	//writeFileSCC(wstream,obj)
	writeFileSCC(wstream,data)
	wstream.end();
	
}//writeSCC
function writeFileSCC(strm, data){
// ********
//Scenarist_SCC V1.0

//00:00:00:00	9425 94AD 9170 3E3E 20D3 70E5 616B E96E 6720 EFE6 20E3 6861 EFF4 E9E3 2C80
//00:00:00:29	9425 94AD 9170 7061 F2F4 7320 EFE6 20C7 E5EF F267 E961 2061 6E64 20F4 68E5
//00:00:02:07	9425 94AD 9170 4361 F2EF ECE9 6E61 7320 61F2 E520 64E5 61EC E96E 6720 F7E9 F468 20F4 68E9 73BA
//00:00:03:27	9425 94AD 9170 F468 E973 20E9 7320 E9E3 E520 E6EF F26D E96E 6720 EF6E 20F4 F2E5 E573 AE80
//00:00:05:18	9425 94AD 9170 D9EF 7520 E361 6E20 73E5 E520 F468 E520 E9E3 E9E3 ECE5 7380
//00:00:06:28	9425 94AD 9170 F468 E56D 73E5 EC76 E573 20EA 7573 F420 E6EF F26D E96E 6720 61EC EF6E 6780
//00:00:08:17	9425 94AD 9170 6775 F4F4 E5F2 73AE
//00:00:09:02	9425 94AD 9170 5468 E579 A7F2 E520 E96E 20F4 68E5 2064 61F2 6B20 F468 E973 206D EFF2 6EE9 6E67
// ********

	//write header section
	strm.write('Scenarist_SCC V1.0\n');
	strm.write('\n');
	
	/*
	var intime, outtime;
	for( var i=0; i<o.events.length; i++ ){
		intime = o.events[i].intime;
		outtime = o.events[i].outtime
		var lines = o.
		for( var j=0; i<o.events.length; i++ ){
			strm.write(o.events[i].intime.toTimecodeSMPTE() + '\t' +'something here'+ '\n');
		}
	}
	*/
	
	//PAC array
	var RCL = String.fromCharCode(0x14) + String.fromCharCode(0x20); //"1420 ";//Resume Caption Loading (pop-on)
	var BS  = String.fromCharCode(0x14) + String.fromCharCode(0x21); //"1421 ";//Backspace
	var AOF = String.fromCharCode(0x14) + String.fromCharCode(0x22); //"1422 ";//reserved (formerly Alarm Off)
	var AON = String.fromCharCode(0x14) + String.fromCharCode(0x23); //"1423 ";//reserved (formerly Alarm On)
	var DER = String.fromCharCode(0x14) + String.fromCharCode(0x24); //"1424 ";//Delete to End of Row
	var RU2 = String.fromCharCode(0x14) + String.fromCharCode(0x25); //"1425 ";//Rollup Captions 2-Rows
	var RU3 = String.fromCharCode(0x14) + String.fromCharCode(0x26); //"1426 ";//Rollup Captions 3-Rows
	var RU4 = String.fromCharCode(0x14) + String.fromCharCode(0x27); //"1427 ";//Rollup Captions 4-Rows
	var FON = String.fromCharCode(0x14) + String.fromCharCode(0x28); //"1428 ";//Flash On
	var RDC = String.fromCharCode(0x14) + String.fromCharCode(0x29); //"1429 ";//Resume Direct Captioning (paint-on)
	var TR  = String.fromCharCode(0x14) + String.fromCharCode(0x2A); //"142A ";//Text Restart
	var RTD = String.fromCharCode(0x14) + String.fromCharCode(0x2B); //"142B ";//Resume Text Display
	var EDM = String.fromCharCode(0x14) + String.fromCharCode(0x2C); //"142C ";//Erase Displayed Memory
	var CR =  String.fromCharCode(0x14) + String.fromCharCode(0x2D); //"142D ";//Carriage Return (roll display)
	var ENM = String.fromCharCode(0x14) + String.fromCharCode(0x2E); //"142E ";//Erase Non-Displayed Memory
	var EOC = String.fromCharCode(0x14) + String.fromCharCode(0x2F); //"142F ";//End of Caption (Flip Memories)
	var CLEARINGPULSE = RCL + EDM;
	var PAC = [ 
				[0x11,0x50],[0x11,0x70],
				[0x12,0x50],[0x12,0x70],
				[0x15,0x50],[0x15,0x70],
				[0x16,0x50],[0x16,0x70],
				[0x17,0x50],[0x17,0x70],
				[0x10,0x50],[0x13,0x50],
				[0x13,0x70],[0x14,0x50],
				[0x14,0x70]
			];
	var ceastyle = data.style;//CEA_DEFAULT_STYLE;
	var ceadepth = data.depth;//CEA_DEFAULT_DEPTH;
	var nextavailablebursttime=0;
	for( var i=0; i<data.getEvents().length; i++ ){
		var ev = data.getEvents()[i];
		var lines = ev.getLines();
		var lineobj = {}
		
		//fix up explicit outtimes; CADET sends it to server with all times filled out; intuit blank outtimes...
		if( i < data.getEvents().length-1 ){
			var foo1 = ev.intime;
			var foo2 = ev.outtime;
			var foo3 = data.getEvents()[i+1].intime;
			var foo4 = foo3 - foo2;
			//console.log("writeFileSCC: ev#=["+i+"] original: in=["+foo1+"] out=["+foo2+"] nextin=["+foo3+"] dif=["+foo4+"] check for blinktime");
			
			if( (foo4) <= 0.02 ){
				ev.outtime = -1;
			}
		}
		//console.log("writeFileSCC: ev#=["+i+"] after fixup: in=["+ev.intime+"] out=["+ev.outtime+"]");
		
		var burst;
		
		//CCSTYLE: POP-ON->RESUME CAPTION LOADING and ERASE NON-DISPLAYED MEMORY; ROLLUP->RU#
		if( ceastyle == CEA_STYLE_POPON ){
			burst = RCL + ENM;
		} else {  //assume rollup
			switch( ceadepth ){
				case 2:
					burst = RU2;
					break;
				case 3:
					burst = RU3;
					break;
				case 4:
					burst = RU4;
					break;
			}
			burst += CR;	//roll it up before adding text so text lingers as long as possible
		}
		
		//figure out length of longest line in the event
		var longestlinelength=0;
		for( var k=0; k<lines.length; k++ ){
			if( lines[k].trim().length > longestlinelength ) longestlinelength = lines[k].trim().length;
		}
		//console.log("writeFileSCC: longestlinelength=["+longestlinelength+"]");
		
		
		//determine row and starting column for each line
		for( var j=0; j<lines.length; j++ ){

			lineobj.text = lines[j].trim();
		
			//row
			if( ceastyle == CEA_STYLE_POPON ){
				switch( ev.valign ){
					case ALIGN_TOP:
						lineobj.row = j;
						break;
					case ALIGN_BOTTOM:
						lineobj.row = 15-lines.length+j; //each line moves up from the bottom
						break;
					default:
						lineobj.row = 15-lines.length+j; //assume bottom
				}
			} else { //assume rollup, this is base row
				switch( ev.valign ){
					case ALIGN_TOP:
						lineobj.row = ceadepth-1; //depth, but zero-based
						break;
					case ALIGN_BOTTOM:
						lineobj.row = 14; //bottom
						break;
					default:
						lineobj.row = 14; //assume bottom
				}
			}
			
			//column
			if( ceastyle == CEA_STYLE_POPON ){
				switch( ev.align ){
					case ALIGN_LEFT:
						lineobj.col = 0;
						break;
					case ALIGN_CENTER:
						lineobj.col = 16-(Math.floor(lineobj.text.length/2));
						break;
					case ALIGN_RIGHT:
						//based on longest line
						lineobj.col = 32-longestlinelength;
						break;
					default:
						col = -1; //assume center alignment
				}
			} else { //assume rollup
				lineobj.col = 0;
			}
			
			//compute tab and offset for the line
			lineobj.tab = Math.floor( lineobj.col / 4 );
			lineobj.taboffset = lineobj.col % 4;
			
			//console.log("lineobj=**"+JSON.stringify(lineobj)+"**");
			
			//PAC-ROW
			sfoo = String.fromCharCode(PAC[lineobj.row][0]);
			burst += sfoo
			//PAC-COL
			sfoo = String.fromCharCode(PAC[lineobj.row][1] + (lineobj.tab*2));
			burst += sfoo
			//taboffset
			if( lineobj.taboffset > 0 ){
				sfoo = String.fromCharCode(0x17) + String.fromCharCode(0x20+lineobj.taboffset);
				burst += sfoo
			}
			
			//textline: make sure it's an even number of characters, pad with null if necessary
			var thechar
			for( var k=0; k<lineobj.text.length; k++ ){
				//var achar = lineobj.text.substr(k,1);
				//console.log("achar=["+achar+"] val=["+achar.charCodeAt(0)+"]");
				thechar = ceaMapChar(lineobj.text.substr(k,1));
				if( thechar.length == 1 ){
					//console.log("writeFileSCC: mapped length == 1");
					burst += thechar;
				} else { //make sure to frame align the pair
					//console.log("writeFileSCC: mapped length != 1");
					if( burst.length % 2 == 0 ){
						burst += thechar;
					} else {
						burst += (String.fromCharCode(0) + thechar);
					}
				}
			}

			
			if( burst.length %2 == 1 ){
				burst += String.fromCharCode(0);
			}
			
		}//each line in the event
		
		
		//if popon then add the EOC
		//if( ceastyle ){
			var bBacktime = false; //default for rollup, don't backtime, just start the burst at the intime
			if( data.style == CEA_STYLE_POPON){
				burst += (EDM+EOC); //make it blink before displaying
				bBacktime = true;
			} else {
			}
		//} else {	//assume popon
		//		burst += (EDM+EOC); //make it blink before displaying
		//}
		
		//backtime the intime for buildup. Two bytes per frame, each frame is 1/30 sec = 0.0333s
		var origtime, backtime, outtime;
		origtime = ev.intime;
		outtime = ev.outtime;
		backtime = origtime - (burst.length / 2 * 0.0333 );
		
		if( backtime < 0 ) backtime = 0;
		
		var intimetouse;
		//which one to use
		if( bBacktime ){
			intimetouse = backtime;
		} else {
			intimetouse = origtime;
		}
		
		//make sure burst is not starting before the last one finished
		
		if( intimetouse < nextavailablebursttime ){
			intimetouse = nextavailablebursttime;
		}
			
		//console.log("origtime=["+origtime+"] backtime=["+backtime+"] backtimeSMPTE=["+backtime.toTimecodeSMPTE()+"]");

		
		var str = burst;
		var parts = str.match(/[\s\S]{1,2}/g) || []; //pairs of characters
		
		/*
		//prepend a clearing pulse for the first row if there's time. This helps QuickTime player get things right
		if( i == 0 ){
			if( intimetouse > 0.133 ){
				//00:00:00;01	9420 942c //erase displayed memory
				parts = parts.concat(CLEARINGPULSE.match(/[\s\S]{1,2}/g));
			}
		}
		*/
		nextavailablebursttime = intimetouse + (parts.length * 0.0333) + 0.0333; //one frame past the end of the burst
		
		var charpair;
		var hexburst = "";
		for( var k=0; k<parts.length; k++ ){
			charpair = parts[k];
			if( charpair.length = 2 )
				hexburst += (forceOddParity(charpair.charCodeAt(0)).toHexString() + forceOddParity(charpair.charCodeAt(1)).toHexString() + " ")
			else
				hexburst += (forceOddParity(charpair.charCodeAt(0)).toHexString() + forceOddParity(0x0).toHexString() + " ");//add a null to the end
		}
		
		/*
		*/
		//insert a clearing pulse at 00:00:00;00 if there's room. This helps QuickTime player get things right
		if( i == 0 ){
			if( intimetouse > 0.133 ){
				//00:00:00;01	9420 942c //erase displayed memory
				strm.write( "00:00:00;01" + '\t' + "9420 942c" + '\n');
			}
		}
		strm.write( intimetouse.toTimecodeSMPTE() + '\t' + hexburst + '\n');

		nextavailablebursttime = intimetouse + (parts.length * 0.0333) + 0.0333; //one frame past the end of the burst

		
		//write an erasure event if there's an outtime
		//console.log("writeFileSCC: check for i=["+i+"] outtime=["+outtime.toTimecodeSMPTE()+"]");
		if( outtime > 0 ){
		
			var outtimetouse = outtime-0.1;
			//make sure burst is not starting before the last one finished
			if( outtime-.1 < nextavailablebursttime ){
				outtimetouse = nextavailablebursttime;
			}
		
			//hh:mm:ss;ff	9420 942c //erase displayed memory
			//strm.write( (outtime-.1).toTimecodeSMPTE() + '\t' + "9420 942c" + '\n');
			strm.write( outtimetouse.toTimecodeSMPTE() + '\t' + "9420 942c" + '\n');
			
			nextavailablebursttime = outtimetouse + (2 * 0.0333) + 0.0333; //one frame past the end of the burst
		}
		
	}// each i'th event
	//write last outtime
	//strm.write(ev.outtime.toTimecodeSMPTE() + '\t' +'something here'+ '\n');
	
}//writeFileSCC

var ceaEleven30To38 = "®°½¨™¢£♪à"   //H30 to H38
var ceaEleven3ATo3F = "èâêîôû"   //H3A to H3F
var ceaTwelve20To3F = "ÁÉÓÚÜü‘¡*'—©?•\"ÀÂÇÈÊËëÎÏïÔÙùÛ«»";
var ceaThirteen20To3F = "ÃãÍÌìÒòÕõ{}\^_¦~ÄäÖöß¥¤|ÅåØø????";
function ceaMapChar( s ){

	if( s.charCodeAt(0) >= 0x20 && s.charCodeAt(0) <= 0x7E )
		return(s);

	//otherwise check the special maps
	var n;
	
	//special chars
	n = ceaEleven30To38.indexOf(s);
	if( n > -1 ){
		//console.log
		return (String.fromCharCode(0x11) + String.fromCharCode(0x30+n));
	}
	n = ceaEleven3ATo3F.indexOf(s);
	if( n > -1 ){
		//console.log
		return (String.fromCharCode(0x11) + String.fromCharCode(0x3A+n));
	}
		
	n = ceaTwelve20To3F.indexOf(s);
	if( n > -1 ){
		return (String.fromCharCode(0x12) + String.fromCharCode(0x20+n));
	}
		
	n = ceaThirteen20To3F.indexOf(s);
	if( n > -1 ){
		return (String.fromCharCode(0x13) + String.fromCharCode(0x20+n));
	}
		
	//otherwise just return the original char
	return s;
}//ceaMapChar
function forceOddParity(x){
	var xorig = x;

	x ^= x >> 4;
	x ^= x >> 2;
	x ^= x >> 1;

	//return (~x) & 1;
	if( (~x) & 1 ){
		//console.log("forceOdd: in=["+xorig.toHexString()+"] out=["+(xorig | 0x80).toHexString()+"]");
		return (xorig | 0x80)
	} else {
		//console.log("forceOdd: in=["+xorig.toHexString()+"] out=["+xorig.toHexString()+"]");
		return xorig
	}
}//forceOddParity
function eventToSCC(e){
	var sfoo="", ret="", tr={};
	for( var i=0; i<e.textRuns.length; i++ ){
		tr=e.textRuns[i];
		sfoo = tr.text;
		//log.debug("eventToQTtext: sfoo=["+sfoo+"]");
		if( sfoo.length==0 ) sfoo = " ";
		
		//prepend newline for this run
		if( tr.linebreak ) sfoo = "\n"; //sfoo = "\n"+sfoo;

		ret += sfoo;
	}//textRuns
	
	return ret;
}//eventToSCC


///////////////////////
// EXPORT PLAIN TEXT
///////////////////////
/* blank-line-delimited captions

caption here

another caption here


*/
function writePlainText(fpath, obj){
	//log.debug("writePlainText: enter");
	var wstream = fs.createWriteStream(fpath, {encoding: "utf8"});
	writeFilePlainText(wstream,obj)
	wstream.end();
}//writePlainText
function writeFilePlainText(strm, o){
	//log.debug("writeFilePlainText: enter");

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//write textruns
		strm.write(os.EOL + eventToPlainText(ev) + os.EOL);
	}
	
}//writeFilePlainText
function eventToPlainText(e){
	//log.debug("eventToPlainText: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToPlainText: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += (os.EOL);
			continue;
		} else {
		}
		
		//log.debug("eventToPlainText: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		ret += sfoo;
		//log.debug("eventToPlainText: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToPlainText: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToPlainText


///////////////////////
// EXPORT PLAIN TEXT
///////////////////////
/* blank-line-delimited captions

caption here

another caption here


*/
function writeADText(fpath, obj){
	//log.debug("writePlainText: enter");
	var wstream = fs.createWriteStream(fpath, {encoding: "utf8"});
	writeFileADText(wstream,obj)
	wstream.end();
}//writeADText
function writeFileADText(strm, o){
	//no header
	var stitle = o.config.webProjectFile;
	stitle = stitle.substring(0,stitle.lastIndexOf('.'));
	strm.write('<!doctype html>\n<head>\n<title>'+stitle+'</title>\n</head>\n<body>'); 
	for( var i=0; i<o.events.length; i++ ){
		strm.write('\n<p style="orphans:6; widows:6; width:800px; font-size:1.25rem;">');
		strm.write('EVENT: '+(i+1)+'<br />');
		strm.write('IN: [' + o.events[i].intime.toTimecode(2,false) + ']<br />');
		strm.write('CUE: ' + o.events[i].cueText + '<br />');
		
		strm.write('<b><span style="font-size:2rem;">');
		strm.write(eventToADText(o.events[i]));
		strm.write('</span></b>');
		
		strm.write('<br />OUT: [' + o.events[i].intime.toTimecode(2,false) + ']<br />');
		strm.write('<br /><br /><br /><br /><br />');
		strm.write("</p>");
	}
	strm.write('\n</body>\n</html>'); 
	
	
}//writeFileADText
function eventToADText(e){
	//log.debug("eventToDVSText: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToDVSText: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToDVSText: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += (os.EOL);
			continue;
		} else {
		}
		
		//log.debug("eventToDVSText: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		ret += sfoo;
		//log.debug("eventToPlainText: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToDVSText: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToADText



//project files stored as JSON data
app.post("/openProject/", function(req, res) {
	//log.debug(".post(/openProject/): ENTER");
	//log.debug(".post(/openProject/): stringify data=["+JSON.stringify(req.body)+"]");
	var obj =  JSON.parse(req.body.thedata);
	var projPath = obj.projPath;
	//log.debug(".post(/openProject/): projPath=["+projPath+"]");
	openProject(projPath, res);
})//.post("/openProject/"
function openProject(spath, res){

	/**/
	//log.debug("loadProject: spath=["+spath+"]");
	var projPath = webroot + "/" + spath;
	//log.debug("loadProject: projPath=["+projPath+"]");
	
	fs.readFile( projPath, 'utf8', function (err,data) {
	  if (err) {
	  	log.debug("loadProject: Error reading file=["+projPath+"]");
		return log.debug(err);
	  }
	  //log.debug("loadProject: data=["+data+"]");
	  res.send( data );
	});
	
	
	/*
	var myObject = JSON.parse(fs.createReadStream(projPath));
	strm.read(JSON.stringify(oProject));
	strm.end
	log.debug("loadProject: Stringify=["+JSON.stringify(myObject)+"]");
	*/
	
}//openProject


function rgbToName(c){
	if( c == "#000000" )
		return("black");
	else if( c == "#ffffff" )
		return("white");
	else if( c == "#ff0000" )
		return("red");
	else if( c == "#008000" )
		return("green");
	else if( c == "#0000ff" )
		return("blue");
	else if( c == "#ffff00" )
		return("yellow");
	else if( c == "#00ffff" )
		return("cyan");
	else if( c == "#ff00ff" )
		return("magenta");
	else
		return(c);
}//rgbToName

function itoeAlign(i){
	if( i==ALIGN_LEFT )
		return( "left" );
	else if (i==ALIGN_CENTER )
		return( "center" );
	else if (i==ALIGN_RIGHT )
		return( "right" );
	else
		return( "center" );
}//itoeAlign

function etoiAlign(s){
	if( s=="left" )
		return( ALIGN_LEFT );
	else if (s=="center" )
		return( ALIGN_CENTER );
	else if (s=="right" )
		return( ALIGN_RIGHT );
	else
		return( ALIGN_CENTER );
}//etoiAlign

function itoeVAlign(i){
	if( i==ALIGN_TOP )
		return( "0" );
	else if (i==ALIGN_BOTTOM )
		return( "-1" );
	else
		return( "-1" );
}//itoeVAlign

function etoiVAlign(s){
	if( s=="0" )
		return( ALIGN_TOP );
	else if (s=="-1" )
		return( ALIGN_BOTTOM );
	else
		return( ALIGN_BOTTOM );
}//etoiVAlign

function etoiBold(s){
	if( s=="bold" )
		return( true );
	else
		return( false );
}//etoiBold
function etoiItalic(s){
	if( s=="italic" ||s=="oblique" )
		return( true );
	else
		return( false );
}//etoiItalic
function etoiUnderline(s){
	if( s=="underline" )
		return( true );
	else
		return( false );
}//etoiItalic

function getDateTime() {

    var date = new Date();

    var hour = date.getHours();
    hour = (hour < 10 ? "0" : "") + hour;

    var min  = date.getMinutes();
    min = (min < 10 ? "0" : "") + min;

    var sec  = date.getSeconds();
    sec = (sec < 10 ? "0" : "") + sec;

    var year = date.getFullYear();

    var month = date.getMonth() + 1;
    month = (month < 10 ? "0" : "") + month;

    var day  = date.getDate();
    day = (day < 10 ? "0" : "") + day;

    return year + ":" + month + ":" + day + ":" + hour + ":" + min + ":" + sec;
}//getDateTime



//////////////////////////
// ROUTE IMPORTTYPE
//////////////////////////
app.post("/importtype/:itype", function(req, res) {
    var itype = req.params.itype.toString().toLowerCase();
	//log.debug("app.post(//importtype//): ENTER with itype=["+itype+"]");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);    
    
	//log.debug("app.post(//importtype//): file name=["+myObject.filename+"]");
	//log.debug("app.post(//importtype//): file dir=["+myObject.dir+"]");
	
    var filepath;
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    if( !myObject.dir && !myObject.filename ){
    	log.debug("app.post(//importtype//):!!! RETURNING because  !myObject.dir || !myObject.filename");
    	return;
    }
    filepath = webroot+myObject.dir + "/" + myObject.filename;
    //log.debug("app.post(//importtype//): filepath=["+filepath+"]");
    
    //log.debug("app.post(//importtype//): import type=["+itype+"]");
	switch (itype.toLowerCase()) {
		case "vtt":
			importFileVTT(filepath, res);
			return;
			break;
		case "srt":
			//console.log("SRT import not implemented");
			//return;
			importFileSRT(filepath, res);
			return;
			break;
		case "ttml":
			importFileTTML(filepath, res);
			return;
			break;
		case "qttext":
			importFileQTTEXT(filepath, res);
			return;
			break;
		case "sami":
			console.log("SAMI import not implemented");
			return;
			break;
		case "plain":
			importFilePlain(filepath, res);
			break;
		case "scc":
			importFileSCC(filepath, res);
			break;
		default:
			console.log("No valid import type specified ["+itype+"].");
			res.status(400).send("Server says Invalid import type specified ["+itype+"]");
			return
	}
    
	return;
});//app.post("/importtype/:itype"


///////////////////////
// IMPORT PLAIN TEXT
///////////////////////
function importFilePlain(filepath, res){

	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
		
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var inacue=false, firsttext=true;
	var thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFilePlain: line=["+line+"]");
		if( line.trim().length > 0 ){
			//console.log("importFilePlain: found text so inside a cue");
			if( !inacue ){//not already in a cue so create new event
				//data.addEvent("00:00:00.000".timecodeToNumber(),"24:00:00.000".timecodeToNumber());
				data.addEvent("".timecodeToNumber(),"".timecodeToNumber());
				inacue = true;
				firsttext = true;
			}
			
			thetext = line;
			if( firsttext ){
				//console.log("importFilePlain: first text line of cue = ["+thetext+"]");
				
				cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
				textRun = new CadetTextRun( cadetStyle, thetext );
				data.addRun(textRun, false);	//newline indicated as false

				firsttext = false;
				
			} else {	//2nd or more line
				//console.log("importFilePlain: additional text line of cue = ["+thetext+"]");
				//first add blank run with newline
				cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
				textRun = new CadetTextRun( cadetStyle, "" );
				data.addRun(textRun, false);	//newline indicated as true
				//now add the actual text line
				cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
				textRun = new CadetTextRun( cadetStyle, thetext );
				data.addRun(textRun, false);	//newline indicated as true

			}
		} else {
			//console.log("importFilePlain: blank line, end of cue");
			inacue = false;
		}
			
		if (last) {
			var thedata={};
			thedata.data = data;
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importFilePlain ["+filepath+"] should never get here");
	
}//importFilePlain


///////////////////////
// IMPORT SCC
///////////////////////
function importFileSCC (filepath, res){

	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
		
	var lineReader = require('line-reader');
	var bfoundscc = false;
	var inacue=false, sremainder="", firsttext=true;
	var schunk, sintime = "", thetext="";
	var ceastyle = CEA_STYLE_POPON;
	var ceadepth = 2;
	var cearollbase = 14;
	var eventCount = -1;
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFileSCC: line=["+line+"]");
		if( !bfoundscc ){
			var iscenarist;
			iscenarist = line.toLowerCase().indexOf("scenarist_scc v1.0");
			if( iscenarist == -1 ){
				//console.log("importFileSCC: INVALID SCC FILE");
				res.status(400).send("Server says importFileSCC: ["+filepath+"] is not a valid SCC file.");
				return;
			} else {
				bfoundscc = true;
				//console.log("importFileSCC: VALID SCC FILE");
			}
			bfoundscc = true;
			
		} else { //already found SCENARIST_SCC indicator
		
			if( line.trim().length > 0 ){
				//console.log("importFileSCC: found non-blank line=["+line+"]");
			
				//first non-blank line must be  "Scenarist_SCC V1.0"
				if( !bfoundscc ){
					var iscenarist;
					iscenarist = line.toLowerCase().indexOf("scenarist_scc v1.0");
					if( iscenarist == -1 ){
						res.status(400).send("Server says importFileSCC: ["+filepath+"] is not a valid SCC file.");
						return;
					} else {
						bfoundscc = true;
					}
					bfoundscc = true;
				}
			
				// ********
				//Scenarist_SCC V1.0

				//00:00:00;01	9420 9420 942c 942c 942f 942f
				// ********
			
				var spairs, s1, s2, i1, i2, c1, c2;

				//data.addEvent("00:00:00.000".timecodeToNumber(),"24:00:00.000".timecodeToNumber());
				
				//extract timecode
				schunk = line.substr(0, 12);
				sintime_orig = schunk.substr(0, 11); //this is the original timestamp, roughly the intime, but not quite
				sremainder = line.substr(12);
				spairs = sremainder.split(" ");
				
				sintime_num = sintime_orig.timecodeSmpteToNumber();
			
				//console.log("importFileSCC: sintime_orig=["+sintime_orig+"] spairs=["+spairs.length+"] sintime_str=["+sintime_str+"]");
			
				//data.addEvent(sintime_num,"".timecodeToNumber());
				inacue = true;
				firsttext = true;
				
				//var CEA608Decoder = require("./CEA608Decoder.js");
				var line21Decoder = require("./CEA608Decoder.js");
				//var line21Decoder = new CEA608Decoder();
				//console.log("server: html DM=["+line21Decoder.itoeAsHtmlDM()+"]");
				//console.log("server: html NM=["+line21Decoder.itoeAsHtmlNM()+"]");

			
				spairs.forEach(function(spair) {
					i1 = parseInt(spair.substr(0,2), 16) & 0x7f
					i2 = parseInt(spair.substr(2,2), 16) & 0x7f
					s1 = i1.toHexString();
					s2 = i2.toHexString();
					c1 = String.fromCharCode(i1);
					c2 = String.fromCharCode(i2);
					
					//console.log("importFileSCC: spair=["+spair+"] s1=["+i1.toString(16)+"] s2=["+i2.toString(16)+"] i1=["+i1+"] i2=["+i2+"] "+((i1>=32 && i2>=32)?"cc=["+String.fromCharCode(i1)+String.fromCharCode(i2)+"]":""));
					//console.log("c1=["+c1+"]");
					//console.log("c2=["+c2+"]");

					if( line21Decoder.decodeLine21(c1) > 0 ){
						//console.log("!!!!! EVENT char 1 !!!!!!");
						//line21Decoder.dumpDM();
					}
					if( line21Decoder.decodeLine21(c2) > 0 ){
						//console.log("!!!!! EVENT char 2 !!!!!!");
						//line21Decoder.dumpDM();
						//console.log("server: DM html=["+line21Decoder.itoeAsHtmlDM()+"]");
						
					}//if found event on #2


				});	//foreach spair	in spairs
				
				//get the caption style
				ceastyle = line21Decoder.getStyle();
				ceadepth = line21Decoder.getDepth();
				cearollbase = line21Decoder.getRollBase()-1;//decoder returns it 1-based, we want it 0-based
				//console.log("style=["+ceastyle+"] depth=["+ceadepth+"]");
				
				//adjust intime based on style (forward timed or not
				if( ceastyle == CEA_STYLE_POPON ){
					//add buildup time to intime; each pair takes one frame, 1 frame = 0.033 secs
					sintime_num += (spairs.length*0.033);
					//console.log("importFileSCC: added bursttime orig intime=["+sintime_orig+"] curtime=["+sintime_num+"]");
				} else{
					//for rollup, just use the SCC burst time as the popon time
					//sintime_str = sintime_num.toTimecode(2);
					//leave it alone
					//console.log("importFileSCC: left bursttime alone curtime=["+sintime_num+"]");
				}
				//add an event at this time
				data.addEvent(sintime_num," ".timecodeToNumber());
				eventCount++;
								
				console.log("importFileSCC: line21Decoder.itoeAsHtmlDM()=["+line21Decoder.itoeAsHtmlDM()+"]");
				var lines = line21Decoder.itoeAsHtmlDM().split("<br/>");
				lines.forEach(function(line){
					thetext = line;
					if( thetext.length == 0 ) thetext = ""; //make sure there's at least a blank space there
					if( firsttext ){
						//console.log("importFileSCC: first text line of cue = ["+thetext+"]");
	
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						//console.log("importFileSCC: add text run at event intime=["+sintime_num+"] text=["+thetext+"]");
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
	
					} else {	//2nd or more line
						//console.log("importFileSCC: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}
				}); //foreach line in lines
				
				if( ceastyle == CEA_STYLE_POPON ){
					var firstrow = line21Decoder.getFirstRow();
					if( firstrow > -1 ){
						firstrow += 1;//decoder is 1-based, we need 0-based
							if( firstrow <= 6 )
								data.getEvents()[eventCount].setVAlign(ALIGN_TOP);
							else
								data.getEvents()[eventCount].setVAlign(ALIGN_BOTTOM);
						
					}
				
				} else { //assume rollup
					if( cearollbase <= 6 )
						data.getEvents()[eventCount].setVAlign(ALIGN_TOP);
					else
						data.getEvents()[eventCount].setVAlign(ALIGN_BOTTOM);
				}

			
			} else {
				//console.log("importFileSCC: blank line, end of cue");
				inacue = false;
			}
			
			if (last) {
				//console.log("importFileSCC: last");
				//fix up outTimes; each outtime is next intime (minus blinktime)
				for(var i=0; i<data.getEvents().length; i++){
					var nextintime, thisintime, thisouttime, nexttext;
					
					thisintime = data.events[i].getIntime();
					thisouttime = data.events[i].getOuttime();
					//console.log("importFileSCC: last: i=["+i+"] in=["+thisintime.toTimecode(2)+"] out=["+thisouttime.toTimecode(2)+"]");
					
					//console.log("importQTTEXT: i=["+i+"]");
					if( i == data.getEvents().length-1 ){ //last item, set to infinity
							//console.log("importQTTEXT: setting to INFINITY");
							data.events[i].setOuttime(INFINITY.timecodeToNumber());
					} else {	//all other items
						nexttext = data.events[i+1].getText();
						//console.log("importQTTEXT: i=["+i+"] checking next text=["+nexttext+"] length=["+nexttext.trim().length+"]");
						if( nexttext.trim().length == 0 ){	//next one is an erasure, so use the time as outtime
							nextintime = data.events[i+1].getIntime();
							//console.log("importQTTEXT: last: i=["+i+"] setting outtime to nextintime=["+nextintime.toTimecode(2)+"]");
							data.events[i].setOuttime(nextintime);//not otherwise specfied
						} else {	//generic, use -1
							//console.log("importQTTEXT: setting outtime to -1");
							data.events[i].setOuttime(-1);//not otherwise specfied
						}
					}
				}//for all items
				/**/
				// now remove elements with blank captions
				for(var i=data.getEvents().length-1; i>=0; i--){
					if( data.getEvents()[i].getText().trim().length == 0 ){
						data.getEvents().splice(i,1);
					}
				}


				var thedata={};
				thedata.data = data;
				thedata.data.ceastyle = ceastyle;
				thedata.data.ceadepth = ceadepth;
				//console.log("after last line of file: style=["+ceastyle+"] depth=["+ceadepth+"]");

				//console.log("FINISHED: thedata="+JSON.stringify(thedata,null,2));
				res.status(200).send(JSON.stringify(thedata));
				return;
			}// if last line of file
			
		}// if bfoundscc
				
	}); //eachline in linereader

	//res.status(400).send("Server says importFileSCC ["+filepath+"] should never get here");
	
}//importFileSCC
/////////////////// 
// IMPORT WEBVTT
///////////////////
// 1
// 00:01:47.430 --> 00:01:51.900   align:center
// <c.white.bgblack>This blade has a dark past.</c>
// 
// 
//var re_time = /(\d+\:\d+\:\d+\.\d+)/;
//var re_time = /-->/;
function importFileVTT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	//var re_time = /(\d+\:\d+\:\d+.\d+)\s+-->\s+(\d+\:\d+\:\d+.\d+)/;
	var re_time = /(\d+\:\d+\:\d+[\.\,]\d+)\s+-->\s+(\d+\:\d+\:\d+[\.\,]\d+)/;
	var re_cuestyle = /(<.*?>)/g;
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFileVTT: line=["+line+"]");
		if( bfirstline ){
			//console.log("importFileVTT: first line detected");
			if( line.toLowerCase().search("webvtt") < 0 ){
				res.status(400).send("File ["+filepath+"] doesn't appear to be a WebVTT file.");
				return( false ); //tell line-reader to stop reading and close
			}
		} else {
			if( inacue ){ //check for line of text, either first or not
				//console.log("importFileVTT: inside a cue");
				if( line.trim().length > 0 ){
					//console.log("importFileVTT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importFileVTT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importFileVTT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}
				} else {
					//console.log("importFileVTT: blank line, end of cue");
					inacue = false;
				}
			} else { //check for timecodes, indicating start of cue
				//console.log("importFileVTT: NOT inside a cue");
				if( re_time.test(line) ){	//found a timecode line
					var args = re_time.exec(line);
					intime = args[1];
					outtime = args[2];
					//console.log("importFileVTT: found timecode line in=["+intime+"] out=["+outtime+"]");
					data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
					//console.log("importFileVTT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
					//(data.getEvents())[data.getEvents().length-1].printTimes();
					//data.printTimes();
					inacue = true;
					firsttext = true; //get ready for first line of cue, probably eating blank lines or other
				} else {
					//console.log("importFileVTT: not in a cue, not a timecode line");
				}
			}
		
		}
		bfirstline = false;
		//return;
	
		//linearray.push(line);
		
		if (last) {
			var thedata={};
			thedata.lines = linearray;
			thedata.data = data;
			//console.log("importFileVTT: eventcount=["+data.getEvents().length+"]");
			//console.log(JSON.stringify(data,null,2));
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importFileVTT ["+filepath+"] should never get here");
}//importFileVTT

///////////////////////
// IMPORT SRT
///////////////////////
// 1
// 00:01:47.430 --> 00:01:51.900   align:center
// <i>This</i> <b>blade</b> has a <u>dark past</u>.
// 
// 
//var re_time = /(\d+\:\d+\:\d+\.\d+)/;
//var re_time = /-->/;
function importFileSRT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	var re_time = /(\d+\:\d+\:\d+[\.\,]\d+)\s+-->\s+(\d+\:\d+\:\d+[\.\,]\d+)/;
	var re_cuestyle = /(<.*?>)/g;
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFileSRT: line=["+line+"]");
			if( inacue ){ //check for line of text, either first or not
				//console.log("importFileSRT: inside a cue");
				if( line.trim().length > 0 ){
					//console.log("importFileSRT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importFileSRT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importFileSRT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}
				} else {
					//console.log("importFileSRT: blank line, end of cue");
					inacue = false;
				}
			} else { //check for timecodes, indicating start of cue
				//console.log("importFileSRT: NOT inside a cue");
				if( re_time.test(line) ){	//found a timecode line
					var args = re_time.exec(line);
					intime = args[1].replace(",",".");
					outtime = args[2].replace(",",".");
					//console.log("importFileSRT: found timecode line in=["+intime+"] out=["+outtime+"]");
					data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
					//console.log("importFileSRT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
					//(data.getEvents())[data.getEvents().length-1].printTimes();
					//data.printTimes();
					inacue = true;
					firsttext = true; //get ready for first line of cue, probably eating blank lines or other
				} else {
					//console.log("importFileSRT: not in a cue, not a timecode line");
				}
			}
		
		
		if (last) {
			var thedata={};
			thedata.data = data;
			//console.log("importFileSRT: eventcount=["+data.getEvents().length+"]");
			//console.log(JSON.stringify(data,null,2));
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});//forEach

	//res.status(400).send("Server says importFileVTT ["+filepath+"] should never get here");
}//importFileSRT


/////////////////
// IMPORT QTTEXT
/////////////////
// 
// https://developer.apple.com/library/mac/documentation/QuickTime/QT6_3/Chap1/QT6WhatsNew.html
// 
// {QTtext}{font:Geneva}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{timeStamps:absolute}{language:0}{textEncoding:0}
// [00:00:00.000]
// Trip to Drumlin Farm
// [00:00:02.000]
// (child noise)
// [00:00:03.000]
// [00:00:04.000]
// Hi kids!
// [00:00:06.000]
// 
function importFileQTTEXT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	var re_time = /(\[)(\d+\:\d+\:\d+\.\d+)(\])/; //this pattern:  [00:00:00.000]
	var re_cuestyle = /(\{.*?\})/g;
	var lineReader = require('line-reader');
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, {encoding: 'utf8'}, function(line, last) {
		//console.log("importQTTEXT: line=["+line+"]");
		if( bfirstline ){//check for the sacred qttext marker
			//console.log("importQTTEXT: first line detected firsttwochars=["+line.charCodeAt(0)+":"+line.charCodeAt(1)+":"+line.charCodeAt(3)+"] line=["+line.toLowerCase()+"]");
			if( line.toLowerCase().search("{qttext}") < 0 ){
				res.status(400).send("File ["+filepath+"] doesn't appear to be a QTtext file.");
				return( false ); //tell line-reader to stop reading and close
			}
			bfirstline = false;
		} else {
			//test for timecode line; if yes, update outtime of current event and create new event
			if( re_time.test(line) ){	//found a timecode line
				inacue = true;
				var args = re_time.exec(line);
				intime = args[2];
				outtime = INFINITY;//args[2];  //fix up the outtimes later
				//console.log("importQTTEXT: found timecode line in=["+intime+"] out=["+outtime+"]");
				data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
				//console.log("importQTTEXT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
				//(data.getEvents())[data.getEvents().length-1].printTimes();
				//data.printTimes();
				firsttext = true; //get ready for first line of cue
			} else if( (line.trim().length > 0) && inacue ){
					//console.log("importQTTEXT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importQTTEXT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importQTTEXT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}//firsttext or not
			} else {
				inacue = false;
			}//timecode or text
		
		}
	
		if (last) {	//fix up all timecodes
		
			//console.log("importQTTEXT: eventcount=["+data.getEvents().length+"]");
			//console.log("importQTTEXT: "+JSON.stringify(data,null,2));
			
			//fix up outTimes; each outtime is next intime (minus blinktime)
			for(var i=0; i<data.getEvents().length; i++){
				var nextintime;
				//console.log("importQTTEXT: i=["+i+"]");
				if( i == data.getEvents().length-1 ){ //last item, set to infinity
						//console.log("importQTTEXT: setting to INFINITY");
						data.events[i].setOuttime(INFINITY.timecodeToNumber());
				} else {	//all other items
					//console.log("importQTTEXT: checking next text length");
					if( data.getText(i+1).trim().length == 0 ){	//next one is an erasure, so use the time as outtime
						nextintime = data.events[i+1].getIntime();
						//console.log("importQTTEXT: setting outtime to nextintime=["+nextintime+"]");
						data.events[i].setOuttime(nextintime);//not otherwise specfied
					} else {	//generic, use -1
						//console.log("importQTTEXT: setting outtime to -1");
						data.events[i].setOuttime(-1);//not otherwise specfied
					}
				}
			}//for all items
			/**/
			// now remove elements with blank captions
			for(var i=data.getEvents().length-1; i>=0; i--){
				if( data.getEvents()[i].getText().trim().length == 0 ){
					data.getEvents().splice(i,1);
				}
			}
			
		
			var thedata={};
			//thedata.lines = linearray;
			thedata.data = data;
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importQTTEXT ["+filepath+"] should never get here");
}//importFileQTTEXT


/////////////////
// IMPORT TTML
/////////////////
// <?xml version="1.0" encoding="utf-8"?>
// <tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">
// 	<head>
// 		<styling>
// 			<style id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />
// 		</styling>
// 	</head>
// 	<body id="thebody" style="defaultCaption">
// 		<div xml:lang="en">
// 			<p id="1" begin="00:01:47.43" end="00:01:51.90" tts:textAlign="center"><span tts:color="red" tts:backgroundColor="black">This blade</span> has a <span tts:color="white" tts:backgroundColor="black">dark past.</span></p>
// 			<p id="2" begin="00:01:51.91" end="00:01:57.99" tts:textAlign="center">It has shed<br>much innocent blood.</p>
// 			<p id="3" begin="00:01:58.00" end="00:02:01.72" tts:textAlign="center">You're a fool for traveling alone<br/>so completely unprepared.</p>
// 		</div>
// 	</body>
// </tt>
// 
var CadetData = require("./CadetDataNode.js");
var CadetEvent = require("./CadetEventNode.js");
var CadetTextRun = require("./CadetTextRunNode.js");
var CadetStyle = require("./CadetStyleNode.js");
var ttmldata;
var initialAlign = ALIGN_CENTER;
var initialStyle={bold:false, italic:false, underline:false, color:"white", bgcolor:"black", linebreak:false};
var align=initialAlign;
function importFileTTML(filepath, res){
	//console.log("importFileTTML: ttmldata=["+JSON.stringify(ttmldata)+"]");
	//var ev, myrun, cadetStyle;
	

	//*****************************************
	//read the file into the data object
	/**/
	fs.readFile(filepath, 'utf8', function(err, filedata) {
	
		if( err ){
			console.log("importFileTTML: error on readFile ["+filepath+"]");
			res.status(400).send("Server says importFileTTML ["+filepath+"] failed.");
			return;
		}
		
		//strip the UTF-8 BOM if it's there prior to trying to process
		//this seems to work even though the utf8 BOM is EF BB BF;  FE FF is utf-16(BE)  GO FIGURE!!
		filedata = filedata.replace(/^\uFEFF/, '');
		
		//parse the data object into html-ish
		//but first replace malformed self-closing break tags, which Himalaya doesn't handle properly
		//filedata = filedata.toString().replace(/<br\/>/g, "<br \/>");//replace <br/> with <br />
		
		//replace all tabs first
		//filedata = filedata.replace(/\t+/g, "");//replace tabs

		var json = himalaya.parse(filedata);
		//console.log(util.inspect(json[0].children[1], false, null));// [0].body.children[0].div.children[0].p, false, null));

		//var styleOrig = {bold:false, italic:false, underline:false, color:"white", bgcolor:"black", linebreak:false};
		ttmldata = new CadetData();
		//console.log("importFileTTML: calling crawlHimalayaNodes mm:ss=["+getDateTime()+"]") 
		crawlHimalayaNodes(json[0], initialStyle);
		//console.log("importFileTTML: finished crawlHimalayaNodes ms=["+getDateTime()+"]") 
		
		var thedata={};
		//thedata.lines = linearray;
		thedata.data = ttmldata;
		//console.log("importFileVTT: eventcount=["+data.getEvents().length+"]");
		//console.log(JSON.stringify(data,null,2));
		//res.status(400).send("server reports importFileVTT found some data");
		res.status(200).send(JSON.stringify(thedata));
		return;

	});//readfile....
	//*****************************************	
		
}//importFileTTML
var himalayaFoundPara = false;
var ttmlthelinetext = ""
function crawlHimalayaNodes(n, styleIn){
	//console.log("crawlHimalayaNodes: ttmldata=["+JSON.stringify(ttmldata)+"]");
	//console.log("CRAWL: ENTER");
	//if( n.type && n.type=="Element" ) console.log("CRAWL: ENTER: Element=["+n.tagName+"]");

	var localStyle=extend(true, {},styleIn);//save inbound style to pass to next level; make a deep copy, direct assignment is by reference
			
	if( n.type && (n.type == "Text") && himalayaFoundPara ){
		//console.log("crawlHimalayaNodes--- FOUND Text node and himalayaFoundPara=["+himalayaFoundPara+"]");
		var t = n.content.replace(/\t+/g, "");//replace tabs
		t = t.replace(/\n+/g, "");//replace newlines
		t = t.replace(/\r+/g, "");//replace returns
		if( t.trim().length > 0 ){//ignore newlines and returns
			var thetext = t; //n.content;
			//console.log("crawlHimalayaNodes:--- TEXT: thetext=["+thetext+"]");
			//console.log("crawlHimalayaNodes:--- TEXT: ADD STYLERUN with thetext=["+thetext+"] style=["+JSON.stringify(styleIn)+"]");
			cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, styleIn.linebreak); //newline here is false
			textRun = new CadetTextRun( cadetStyle, thetext );
			ttmldata.addRun(textRun, false);	//newline indicated as false
			ttmlthelinetext += thetext; //accumulate the line text
			//console.log("crawlHimalayaNodes: textnode: thetext=["+thetext+"] ttmlthelinetext=["+ttmlthelinetext+"]");
		}
		//himalayaFoundPara = false;//reset
		return;

	} else if( n.tagName && (n.tagName == "tt") ){
		//console.log("crawlHimalayaNodes--- FOUND TT");
		
	} else if( n.tagName && (n.tagName == "head") ){
		//console.log("crawlHimalayaNodes--- FOUND HEAD");
		
	} else if( n.tagName && (n.tagName == "body") ){
		//console.log("crawlHimalayaNodes--- FOUND BODY");
		
	} else if( n.tagName && (n.tagName == "div") ){
		//console.log("crawlHimalayaNodes--- FOUND DIV");
		
	} else if( n.tagName && (n.tagName == "styling") ){
		//console.log("crawlHimalayaNodes--- FOUND STYLING");
		
	} else if( n.tagName && (n.tagName == "style") ){ //maybe grab a default style???
		//console.log("crawlHimalayaNodes--- FOUND STYLE");
		return;
		
	} else if( n.tagName && (n.tagName == "initial") ){ //maybe grab a default style???
		//console.log("crawlHimalayaNodes--- FOUND INITIAL (style)");
		//initialStyle.
		//initialize localStyle
		return;
		
		
	} else if( n.tagName == "p" ){//new caption grab the timecodes and possible textAlign
		//console.log("crawlHimalayaNodes--- FOUND P");
		himalayaFoundPara = true;//signal the next textType node that it's inside a P tag
		ttmlthelinetext = ""; //reset the accumulated line text
		var sintime="", souttime="", salign="";
		if( n.attributes.begin ) sintime = n.attributes.begin;
		if( n.attributes.end ) souttime = n.attributes.end;
		if( n.attributes["tts:textAlign"] ) salign = n.attributes["tts:textAlign"];
		if( n.attributes["tts:textAlign"]) {align = etoiAlign(salign);}
		//console.log("crawlHimalayaNodes--- ADD EVENT with intime=["+sintime+"] outtime=["+souttime+"] align=["+salign+"]");
		ttmldata.addEvent(sintime.timecodeToNumber(),souttime.timecodeToNumber(), align);
		align = initialAlign;//reset alignment to initial alignment (which *could* be set from <styling><initial></styling>)

		
	} else if( n.tagName && (n.tagName=="br") ){
		if( ttmlthelinetext.trim().length > 0 ){
			//console.log("crawlHimalayaNodes--- FOUND BR");
			//console.log("--- ADD LINEBREAK StyleRun");
			//cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
			cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, true); //newline here is true
			textRun = new CadetTextRun( cadetStyle, "" );//just a blank run with linebreak set
			//ttmldata.addRun(textRun, true);	//newline indicated as true
			ttmldata.addRun(textRun, false);	//we're handling the linebreak run directly, so no need for data.addRun to handle it
			//return;
			ttmlthelinetext = "";
		}
	} else if( n.tagName && (n.tagName=="br/") ){
		if( ttmlthelinetext.trim().length > 0 ){
			//console.log("crawlHimalayaNodes--- FOUND BR/");
			//console.log("--- ADD LINEBREAK StyleRun");
			//cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
			cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, true); //newline here is true
			textRun = new CadetTextRun( cadetStyle, "" );//just a blank run with linebreak set
			//ttmldata.addRun(textRun, true);	//newline indicated as true
			ttmldata.addRun(textRun, false);	//we're handling the linebreak run directly, so no need for data.addRun to handle it
			//return; //but don't return, Himalaya doesn't handle this quite properly
			ttmlthelinetext = "";
		}
	} else if( n.tagName && (n.tagName=="br /") ){
		if( ttmlthelinetext.trim().length > 0 ){
			//console.log("crawlHimalayaNodes--- FOUND BR /");
			//console.log("--- ADD LINEBREAK StyleRun");
			//cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
			cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, true); //newline here is true
			textRun = new CadetTextRun( cadetStyle, "" );//just a blank run with linebreak set
			//ttmldata.addRun(textRun, true);	//newline indicated as true
			ttmldata.addRun(textRun, false);	//we're handling the linebreak run directly, so no need for data.addRun to handle it
			//return; //but don't return, Himalaya doesn't handle this quite properly
			ttmlthelinetext = "";
		}
		
	} else if( n.tagName && (n.tagName == "span") ){//update with new style
		//console.log("crawlHimalayaNodes--- FOUND SPAN");
		if( n.attributes["tts:color"]) {localStyle.color = n.attributes["tts:color"];}
		if( n.attributes["tts:backgroundColor"]) {localStyle.bgcolor = n.attributes["tts:backgroundColor"];}
		if( n.attributes["tts:fontWeight"]) {localStyle.bold = etoiBold(n.attributes["tts:fontWeight"]);}
		if( n.attributes["tts:fontStyle"]) {localStyle.italic = etoiItalic(n.attributes["tts:fontStyle"]);}
		if( n.attributes["tts:textDecoration"]) {localStyle.underline = etoiUnderline(n.attributes["tts:textDecoration"]);}
		//if( n.attributes["tts:textAlign"]) {}//{localStyle.align = etoiAlign(n.attributes["tts:textAlign"]);}
		
	}
	
	//crawl all existing children
	if( n.children ){
		//crawl each child
		n.children.forEach(function(element, index, array){
			crawlHimalayaNodes(element, localStyle);
		});
		if( himalayaFoundPara && n.tagName && n.tagName=="p" )
			himalayaFoundPara = false;
	}//if there are children
	
}//crawlHimalayaNodes

